/*
 ============================================================================
 Name        : indexing.c
 Author      : Oscar Contreras
 Version     :
 Copyright   : Copyright (C) 2015-2016 Oscar Contreras. All rights reserved.
 Description : First order logic theorem prover in C
 ============================================================================
 */

/****************************************************************
*
*             indexing (term indexing functions module)
*
* All of the documentation and software included in the Drodi package
* is copyrighted by Oscar Contreras.
*
* Copyright (C) 2015-2016 Oscar Contreras. All rights reserved.
*
*
* This software is provided "as-is," without any express or implied warranty.
* In no event shall the author be held liable for any damages arising from the
* use, or the result of use, of this software and its documentation in terms
* of correctness, accuracy, reliability, currentness, or otherwise.
*
* Permission is granted to anyone to use this software for any purpose,
* including commercial use, and to alter and redistribute it, provided
* that the following conditions are met:
*
* 1. All redistributions of source code files must retain all copyright
*    notices that are currently in place, and this list of conditions without
*    modification.
*
* 2. All redistributions in binary form must retain all occurrences of the
*    above copyright notice that are currently in place (for example, in
*    the About boxes).
*
* 3. The origin of this software must not be misrepresented; you must not
*    claim that you wrote the original software.
*
* 4. Modified versions in source or binary form must be plainly marked as
*    such, and must not be misrepresented as being the original software.
*
*
* Oscar Contreras
* rwtrainer10 AT gmail.com
*
****************************************************************/

/*-------------------------------------------------------------
 *
 *  DESCRIPTION:
 *
 *    This source contains the hashing and term indexing functions
 *    of the Drodi package.
 *
 *    Drodi is a multi-platform Open Source Artificial Intelligence application.
 *    See Drodi.c for additional information.
 *
 *
 *  IMPLEMENTATION:
 *
 *    Multi-platform
 *
 *  NOTES:
 *
 *
 *--------------------------------------------------------------*/

/* Include for all C sources */
#include "global.h"

/** Global variables for this module: only those that need initialization in their definition. */

/*--------------------------------------------------------------
  *
  *  DESCRIPTION: (Add locally defined symbols to global symbol list)  OCJ
  *
  *    This function adds locally defined symbols to global symbol list.
  *
  *
  *  ARGUMENTS:
  *
  *    names: structure with local symbol and variable list.
  *
  *  RETURNS:
  *
  *    0 if OK or NOMEMORY.
  *
  *--------------------------------------------------------------*/
int32_t AddGlobalSymbols(nameschain *names){
	auto int32_t *ptr1;                             /* Auxiliary pointer */
	auto char *auxchar1,*auxchar2;                  /* Auxiliary pointers */
 	auto int32_t ii,jj,kk,mm,nn,oo;                 /* Auxiliary */

 	/* Return if no names buffer allocated. */
 	if (names->names==NULL){
 		return(0);
 	}

 	/* Loop for each symbol. */
 	for (ii=mm=nn=oo=0;names->names[ii]!=0;ii+=(sizeof(int32_t)+jj+1)){

 		/* Check collision with skolem prefixes. */
 		if ((0==strncmp(skfprefix,&names->names[ii],2))&&(NULL!=(auxchar1=strchr(&names->names[ii+2],'_')))){
 			kk=strtol(&names->names[ii+2],&auxchar2,10);
 			if ((auxchar1==auxchar2)&&(kk>=skolemid)){
 				mm=1;
 				skolemid=kk+1;
 			}
 		}

 		/* Check collision with predicate definitions prefix. */
 		if ((0==strncmp(prdprefix,&names->names[ii],2))&&(NULL!=(auxchar1=strchr(&names->names[ii+2],'_')))){
 			kk=strtol(&names->names[ii+2],&auxchar2,10);
 			if ((auxchar1==auxchar2)&&(kk>=splid)){
 				oo=1;
 				pdefid=kk+1;
 			}
 		}

 		/* Check collision with P-predicate prefix. */
 		if ((0==strncmp(splprefix,&names->names[ii],3))&&(NULL!=(auxchar1=strchr(&names->names[ii+3],'_')))){
 			kk=strtol(&names->names[ii+3],&auxchar2,10);
 			if ((auxchar1==auxchar2)&&(kk>=splid)){
 				nn=1;
 				splid=kk+1;
 			}
 		}

 		/* Calculate auxiliary data. */
 		jj=strlen(&names->names[ii]);
 		ptr1=(int32_t *)&names->names[ii+jj+1];

 		/* If name is a predicate or function symbol then */
 		/* create a symbol entry. */
 		if (0==((*ptr1)&0x40000000)){
 			if (NULL==CreateGlobalSymbol(&names->names[ii],(*ptr1)&0x1fffffff,
 					(((*ptr1)&0x20000000)?INCONJECTURE:0)|(((*ptr1)&0x80000000)?FUNCTION:PREDICATE))){
 				return(NOMEMORY);
 			}
 		}
 	}

 	/* Update skolem and/or P-predicate prefixes if necessary. */
 	if (mm){
		skfprefix[2]=0;
		sprintf(&skfprefix[2],"%d_",skolemid);
 	}
 	if (oo){
		prdprefix[2]=0;
		sprintf(&prdprefix[2],"%d_",pdefid);
 	}
 	if (nn){
		splprefix[3]=0;
		sprintf(&splprefix[3],"%d_",splid);
 	}

 	return(0);
 } /* AddGlobalSymbols */

/*--------------------------------------------------------------
  *
  *  DESCRIPTION: (Create global symbol entry)  OCJ
  *
  *    This function creates a global symbol entry in the kb_hash chain.
  *
  *
  *  ARGUMENTS:
  *
  *    name: Pointer to symbol name.
  *    arity: Number of arguments of symbol.
  *    type: Type of symbol. Allowed values are:
  *          FUNCTION: The symbol is a function.
  *          PREDICATE: The symbol is a predicate.
  *          Only one of the flags FUNCTION or PREDICATE can be set.
  *          DEFINEDPRED: The symbol is a defined predicate. This flag
  *          can be set only if the PREDICATE flag is set.
  *          INCONJECTURE: The symbol is in the conjecture.
  *
  *  RETURNS:
  *
  *    Pointer to hashchain symbol structure or NULL if NOMEMORY.
  *
  *--------------------------------------------------------------*/
hashchain *CreateGlobalSymbol(char *name,int32_t arity,int32_t type){
	auto hashchain *ptr2;                           /* Auxiliary pointer */
 	auto int32_t hh,kk;                             /* Auxiliary */

	/* Allocate memory for hash symbol element. */
	kk=sizeof(hashchain)+(SPACE_AFTER_HASHCHAIN*sizeof(void *));
	if (NULL==(ptr2=mallocmn(strlen(name)+kk+1))){
		return(NULL);
	}

	/* Get hash value. Store it in hashchain structure for */
	/* future use in calculating clause components hashes */
	/* when generating P-predicate names. */
	#ifdef MURMUR3
	hh=ptr2->hash=Murmur3(name);
	#else
	hh=ptr2->hash=KRMultHash(name);
	#endif
	hh&=0xffff;

	/* Chain the element to global queue */
	/* and chain the element hash queue.*/
	omp_set_lock(&procctl.omp_lockc1);
	ptr2->nextelem=kb_hash;
	kb_hash=ptr2;
	ptr2->nextsymbol=hashkeys[hh];
	hashkeys[hh]=ptr2;
	omp_unset_lock(&procctl.omp_lockc1);

	/* Set symbol weights for clause selection. */
	if (type&INCONJECTURE){
		if (type&FUNCTION){
			ptr2->weight[0]=ptr2->weight[1]=ptr2->weight[2]=FUNCWEIGHT;
		} else {
			ptr2->weight[0]=ptr2->weight[1]=ptr2->weight[2]=PREDWEIGHT;
		}
	} else {
		if (type&FUNCTION){
			ptr2->weight[0]=FUNCWEIGHT;
			ptr2->weight[1]=FUNCWEIGHT*FNOTINCONJFACTOR1;
			ptr2->weight[2]=FUNCWEIGHT*FNOTINCONJFACTOR2;
		} else {
			ptr2->weight[0]=PREDWEIGHT;
			ptr2->weight[1]=PREDWEIGHT*PNOTINCONJFACTOR1;
			ptr2->weight[2]=PREDWEIGHT*PNOTINCONJFACTOR2;
		}
	}

	/* Set remaining element fields. */
	ptr2->arity=arity;
	ptr2->type=type&(~INCONJECTURE);
	if (name[0]=='"'){
		ptr2->type|=DISTINCTOBJECT;
	}
	ptr2->discr1=(uint8_t **)&ptr2[1];
	ptr2->discr2=&ptr2->discr1[1+num_wkbs];
	ptr2->discr3=&ptr2->discr2[1+num_wkbs];
	ptr2->discr4=&ptr2->discr3[1+num_wkbs];
	ptr2->discr5=&ptr2->discr4[1+num_wkbs];
	ptr2->discr6=&ptr2->discr5[1+num_wkbs];
	numsymbols++;
	ptr2->precedence=numsymbols;
	for (hh=0;hh<=num_wkbs;hh++){
		ptr2->discr1[hh]=ptr2->discr2[hh]=ptr2->discr3[hh]=ptr2->discr4[hh]=
				ptr2->discr5[hh]=ptr2->discr6[hh]=NULL;
	}
	strcpy(((char *)ptr2)+kk,name);


 	return(ptr2);
 } /* CreateGlobalSymbol */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Murmur3 hash function)  OCJ
 *
 *    This function implements the murmur3 hash function. This has been
 *    taken from Wikipedia murmur3 32 bits sample code.
 *
 *  ARGUMENTS:
 *
 *    key: Key to be hashed. It must be a null ended alphanumeric char string.
 *
 *  RETURNS:
 *
 *    Hashed value.
 *
 *--------------------------------------------------------------*/
#ifdef MURMUR3
uint32_t Murmur3(char *key){
	static const uint32_t c1=0xcc9e2d51;
	static const uint32_t c2=0x1b873593;
	static const uint32_t r1=15;
	static const uint32_t r2=13;
	static const uint32_t mm=5;
	static const uint32_t nn=0xe6546b64;
	static const uint32_t seed=0x9747b28c;
	auto uint32_t len;                                /* Key length */
	auto int32_t nblocks;                             /* Number of 4 bytes blocks in key */
	auto uint32_t *blocks;                            /* Pointer to blocks in key */
	auto uint32_t hash;                               /* 32 bit hash value */
	auto unsigned char *tail;                         /* Pointer to remaining byte in key */
	auto int32_t ii;                                  /* Auxiliary */
	auto uint32_t kk,k1;                              /* Auxiliary */

	/* Initialize and modify hash value for each whole chunk of key. */
	hash=seed;
	len=strlen((char *)key);
	nblocks=len/4;
	blocks=(uint32_t *)key;
	for (ii=0;ii<nblocks;ii++) {
		kk=blocks[ii];
		kk*=c1;
		kk=ROT32(kk,r1);
		kk*=c2;
		hash^=kk;
		hash=ROT32(hash,r2)*mm+nn;
	}

	/* Modify hash value for remaining bytes in key. */
	tail=(unsigned char *)(key+nblocks*4);
	k1=0;
	switch (len&3){
		case 3:
			k1^=tail[2]<<16;
			/* no break */
		case 2:
			k1^=tail[1]<<8;
			/* no break */
		case 1:
			k1^=tail[0];
			k1*=c1;
			k1=ROT32(k1,r1);
			k1*=c2;
			hash^=k1;
	}

	/* Final hash value modifications. */
	hash^=len;
	hash^=(hash>>16);
	hash*=0x85ebca6b;
	hash^=(hash >>13);
	hash*=0xc2b2ae35;
	hash^=(hash>>16);

	return(hash);
} /* Murmur3 */

#else
/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Kernighan and Ritchie hash function)  OCJ
 *
 *    This function implements the Kernighan and Ritchie modified
 *    hash function.
 *
 *  ARGUMENTS:
 *
 *    key: Key to be hashed. It must be a null ended alphanumeric
 *         char string.
 *
 *  RETURNS:
 *
 *    Hashed value.
 *
 *--------------------------------------------------------------*/
uint32_t KRMultHash(char *key){
	auto uint32_t hash;                               /* 32 bit hash value */
	auto uint32_t len;                                /* Key length */
	auto int32_t ii;                                  /* Auxiliary */
	hash=0;
	len=strlen((char *)key);
	for(ii=0;ii<len;++ii){
		hash=(hash<<5)-hash+key[ii]-48;
	}
	return(hash);
} /* KRMultHash */
#endif

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Check if name is globally declared)  OCJ
 *
 *    This function checks if a name is declared.
 *
 *
 *  ARGUMENTS:
 *
 *    name: Symbol name to check.
 *
 *  RETURNS:
 *
 *    Pointer to hashchain structure of the symbol
 *    or NULL if symbol is not globally declared.
 *
 *--------------------------------------------------------------*/
hashchain *CheckGlblDeclared(char *name){
	auto hashchain *ptr1;                           /* Auxiliary pointer */
	auto int32_t hh,ii;                             /* Auxiliary */

	/* Get hash value. */
	#ifdef MURMUR3
	hh=Murmur3(name);
	#else
	hh=KRMultHash(name);
	#endif

	/* Loop through hashchain elements. */
	ii=sizeof(hashchain)+(SPACE_AFTER_HASHCHAIN*sizeof(void *));
	for (ptr1=hashkeys[hh&0xffff];ptr1!=NULL;ptr1=ptr1->nextsymbol){
		if (0==strcmp(((char *)ptr1)+ii,name)){
			return(ptr1);
		}
	}

	return(NULL);
} /* CheckGlblDeclared */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Edit discrimination trees for clause)  OCJ
 *
 *    This function performs the discrimination term indexing needed
 *    for literals and terms in a clause.
 *
 *
 *  ARGUMENTS:
 *
 *    clause: Pointer to clause in binary form.
 *    nkb: KB where the clause is placed. This must be a working KB
 *         but this is not verified.
 *    flag: ADDTREES or DELETETREES
 *
 *  RETURNS:
 *
 *    0 if successful, NOMEMORY if not enough memory.
 *
 *--------------------------------------------------------------*/
int32_t EditDiscrTrees(cmprefix *clause,int32_t nkb,int32_t flag){
	auto uint8_t *ptr1,*ptr3,*ptr4,**ptr5;          /* Auxiliary pointers */
	auto symbol *ptr2;                              /* Auxiliary pointer */
	auto int32_t dd,ii,jj,kk,mm;                    /* Auxiliary */

	/* Check if clause is eligible for performing demodulation from it. */
	if ((clause->part2.bin->literals!=1)||(nkb==num_wkbs)){
		dd=0;
	} else {
		dd=1;
	}

	/* Loop through clause items. */
	ii=jj=kk=0; /* Just to avoid compiler warnings. */
	ptr3=ptr4=NULL; /* Just to avoid compiler warnings. */
	for (ptr1=&clause->part2.bin->formula[0];(*ptr1)!=UNITEND;
			ptr1=NextItem(ptr1,IMMED)){

		/* Item is not a variable. */
		if (0==(VARIABLE&ptr1[0])){

			/* Get equality root terms data. */
			/* With this code the following settings are achieved: */
			/* ptr3 = First equality root term if ptr1 is a subterm of positive */
			/*        or negative equality, NULL otherwise. */
			/* ptr4 = Second equality root term if ptr1 is a subterm of positive */
			/*        or negative equality, NULL otherwise. */
			/* ii = 1 if ptr1 is a positive equality or one of its subterms, */
			/* ii = 0 if ptr1 is a negative equality or one of its subterms, */
			/*      undefined otherwise. */
			/* jj = 1 if ptr1 is a selected literal or one of its subterms, */
			/*      0 otherwise. */
			/* kk = 1 if ptr1 is a selected equality root term, */
			/*      0 if ptr1 is an unselected equality root term, */
			/*      undefined otherwise. */
			switch (ptr1[0]&(EQUALITY|PREDICATE|FUNCTION)){
				case EQUALITY:
					ptr3=NextItem(ptr1,IMMED);
					ptr4=NextItem(ptr3,OVERSUBTERMS);
					if (ptr1[0]&NEGATED){
						ii=0;
					} else {
						ii=1;
					}
					if (ptr1[0]&SELECTED){
						jj=1;
					} else {
						jj=0;
					}
					break;
				case PREDICATE:
					ptr3=ptr4=NULL;
					if (ptr1[0]&SELECTED){
						jj=1;
					} else {
						jj=0;
					}
					break;
				case FUNCTION:
					if ((ptr1==ptr3)||(ptr1==ptr4)){
						if (ptr1[0]&SELECTED){
							kk=1;
						} else {
							kk=0;
						}
					}
					break;
			}

			/* Active clause. */
			ptr2=(symbol *)(ptr1+1);
			if (clause->flags&ACTIVE){

				/* Process depending on kind of item. */
				switch (ptr1[0]&(EQUALITY|PREDICATE|NEGATED|FUNCTION)){

					/* Item is a negated predicate. */
					case (PREDICATE|NEGATED):
						if (ptr1[0]&SELECTED){
							if (flag&ADDTREES){
								if (NOMEMORY==DsTermIndexing(ptr1,&ptr2->symbol->discr2[nkb])){
									return(NOMEMORY);
								}
							} else {
								if (NOMEMORY==DeleteBranch(ptr1,&ptr2->symbol->discr2[nkb])){
									return(NOMEMORY);
								}
							}
						} else {
							if (flag&ADDTREES){
								if (NOMEMORY==DsTermIndexing(ptr1,&ptr2->symbol->discr4[nkb])){
									return(NOMEMORY);
								}
							} else {
								if (NOMEMORY==DeleteBranch(ptr1,&ptr2->symbol->discr4[nkb])){
									return(NOMEMORY);
								}
							}
						}
						break;

					/* Item is a negated equality. */
					case (EQUALITY|NEGATED):
						if (ptr1[0]&SELECTED){
							if (flag&ADDTREES){
								if (NOMEMORY==DsTermIndexing(ptr1,&equkey.discr2[nkb])){
									return(NOMEMORY);
								}
							} else {
								if (NOMEMORY==DeleteBranch(ptr1,&equkey.discr2[nkb])){
									return(NOMEMORY);
								}
							}
						} else {
							if (flag&ADDTREES){
								if (NOMEMORY==DsTermIndexing(ptr1,&equkey.discr4[nkb])){
									return(NOMEMORY);
								}
							} else {
								if (NOMEMORY==DeleteBranch(ptr1,&equkey.discr4[nkb])){
									return(NOMEMORY);
								}
							}
						}
						break;

					/* Item is a positive predicate. */
					case PREDICATE:
						if (ptr1[0]&SELECTED){
							if (flag&ADDTREES){
								if (NOMEMORY==DsTermIndexing(ptr1,&ptr2->symbol->discr1[nkb])){
									return(NOMEMORY);
								}
							} else {
								if (NOMEMORY==DeleteBranch(ptr1,&ptr2->symbol->discr1[nkb])){
									return(NOMEMORY);
								}
							}
						} else {
							if (flag&ADDTREES){
								if (NOMEMORY==DsTermIndexing(ptr1,&ptr2->symbol->discr3[nkb])){
									return(NOMEMORY);
								}
							} else {
								if (NOMEMORY==DeleteBranch(ptr1,&ptr2->symbol->discr3[nkb])){
									return(NOMEMORY);
								}
							}
						}
						break;

					/* Item is a positive equality. */
					case EQUALITY:

						/* Equality is selected. */
						if (ptr1[0]&SELECTED){

							/* Add/delete equality to/from tree. */
							if (flag&ADDTREES){
								if (NOMEMORY==DsTermIndexing(ptr1,&equkey.discr1[nkb])){
									return(NOMEMORY);
								}
							} else {
								if (NOMEMORY==DeleteBranch(ptr1,&equkey.discr1[nkb])){
									return(NOMEMORY);
								}
							}

							/* If equality is selected for paramodulating from it and it has */
							/* a selected root term that is a variable then add/delete equality */
							/* to vartrmidx field of kbase structure. */
							if ((ptr1[0]&PMSELECTED)&&(((VARIABLE|SELECTED)==((VARIABLE|SELECTED)&ptr3[0]))
									||((VARIABLE|SELECTED)==((VARIABLE|SELECTED)&ptr4[0])))){

								/* Addition. */
								if (flag&ADDTREES){

									/* Allocate more space for pointers in leave node if needed. */
									if (kbset[nkb]->vartrmidx.size==kbset[nkb]->vartrmidx.used){
										if (NULL==(ptr5=reallocmn(kbset[nkb]->vartrmidx.symbol,
												(TREEPTR_CHUNK_SIZE+kbset[nkb]->vartrmidx.size)*sizeof(void *)))){
											return(NOMEMORY);
										}
										kbset[nkb]->vartrmidx.size+=TREEPTR_CHUNK_SIZE;
										kbset[nkb]->vartrmidx.symbol=ptr5;
									}

									/* Set a leave node pointer to item. */
									kbset[nkb]->vartrmidx.symbol[kbset[nkb]->vartrmidx.used]=ptr1;
									kbset[nkb]->vartrmidx.used++;

								/* Deletion. Scan for an item match and delete it. */
								} else {
									for (mm=0;mm<kbset[nkb]->vartrmidx.used;mm++){
										if (ptr1==kbset[nkb]->vartrmidx.symbol[mm]){
											kbset[nkb]->vartrmidx.symbol[mm]=
													kbset[nkb]->vartrmidx.symbol[kbset[nkb]->vartrmidx.used-1];
											kbset[nkb]->vartrmidx.used--;
											break;
										}
									}
								}
							}

						/* Equality is not selected. */
						} else {
							if (flag&ADDTREES){
								if (NOMEMORY==DsTermIndexing(ptr1,&equkey.discr3[nkb])){
									return(NOMEMORY);
								}
							} else {
								if (NOMEMORY==DeleteBranch(ptr1,&equkey.discr3[nkb])){
									return(NOMEMORY);
								}
							}
						}
						break;

					/* Item is a function. */
					case FUNCTION:
						if (flag&ADDTREES){
							if (NOMEMORY==DsTermIndexing(ptr1,&ptr2->symbol->discr3[nkb])){
								return(NOMEMORY);
							}
						} else {
							if (NOMEMORY==DeleteBranch(ptr1,&ptr2->symbol->discr3[nkb])){
								return(NOMEMORY);
							}
						}
						if ((ptr1==ptr3)||(ptr1==ptr4)){
							if (ptr1[0]&SELECTED){
								if (ii){
									if (jj){
										if (flag&ADDTREES){
											if (NOMEMORY==DsTermIndexing(ptr1,&ptr2->symbol->discr1[nkb])){
												return(NOMEMORY);
											}
											if (NOMEMORY==DsTermIndexing(ptr1,&ptr2->symbol->discr2[nkb])){
												return(NOMEMORY);
											}
										} else {
											if (NOMEMORY==DeleteBranch(ptr1,&ptr2->symbol->discr1[nkb])){
												return(NOMEMORY);
											}
											if (NOMEMORY==DeleteBranch(ptr1,&ptr2->symbol->discr2[nkb])){
												return(NOMEMORY);
											}
										}
									}
									if (dd){
										if (flag&ADDTREES){
											if (NOMEMORY==DsTermIndexing(ptr1,&ptr2->symbol->discr4[nkb])){
												return(NOMEMORY);
											}
										} else {
											if (NOMEMORY==DeleteBranch(ptr1,&ptr2->symbol->discr4[nkb])){
												return(NOMEMORY);
											}
										}
									}
								} else if (jj){
									if (flag&ADDTREES){
										if (NOMEMORY==DsTermIndexing(ptr1,&ptr2->symbol->discr2[nkb])){
											return(NOMEMORY);
										}
									} else {
										if (NOMEMORY==DeleteBranch(ptr1,&ptr2->symbol->discr2[nkb])){
											return(NOMEMORY);
										}
									}
								}
							}
						} else if (ptr3!=NULL){
							if ((jj)&&(kk)){
								if (flag&ADDTREES){
									if (NOMEMORY==DsTermIndexing(ptr1,&ptr2->symbol->discr2[nkb])){
										return(NOMEMORY);
									}
								} else {
									if (NOMEMORY==DeleteBranch(ptr1,&ptr2->symbol->discr2[nkb])){
										return(NOMEMORY);
									}
								}
							}
						} else if (jj){
							if (flag&ADDTREES){
								if (NOMEMORY==DsTermIndexing(ptr1,&ptr2->symbol->discr2[nkb])){
									return(NOMEMORY);
								}
							} else {
								if (NOMEMORY==DeleteBranch(ptr1,&ptr2->symbol->discr2[nkb])){
									return(NOMEMORY);
								}
							}
						}
						break;
				}

			/* Passive clause. */
			} else {

				/* Process depending on kind of item. */
				switch (ptr1[0]&(EQUALITY|PREDICATE|NEGATED|FUNCTION)){

					/* Item is a negated predicate. */
					case (PREDICATE|NEGATED):
						if (flag&ADDTREES){
							if (NOMEMORY==DsTermIndexing(ptr1,&ptr2->symbol->discr6[nkb])){
								return(NOMEMORY);
							}
						} else {
							if (NOMEMORY==DeleteBranch(ptr1,&ptr2->symbol->discr6[nkb])){
								return(NOMEMORY);
							}
						}
						break;

					/* Item is a negated equality. */
					case (EQUALITY|NEGATED):
						if (flag&ADDTREES){
							if (NOMEMORY==DsTermIndexing(ptr1,&equkey.discr6[nkb])){
								return(NOMEMORY);
							}
						} else {
							if (NOMEMORY==DeleteBranch(ptr1,&equkey.discr6[nkb])){
								return(NOMEMORY);
							}
						}
						break;

					/* Item is a positive predicate. */
					case PREDICATE:
						if (flag&ADDTREES){
							if (NOMEMORY==DsTermIndexing(ptr1,&ptr2->symbol->discr5[nkb])){
								return(NOMEMORY);
							}
						} else {
							if (NOMEMORY==DeleteBranch(ptr1,&ptr2->symbol->discr5[nkb])){
								return(NOMEMORY);
							}
						}
						break;

					/* Item is a positive equality. */
					case EQUALITY:
						if (flag&ADDTREES){
							if (NOMEMORY==DsTermIndexing(ptr1,&equkey.discr5[nkb])){
								return(NOMEMORY);
							}
						} else {
							if (NOMEMORY==DeleteBranch(ptr1,&equkey.discr5[nkb])){
								return(NOMEMORY);
							}
						}
						break;

					/* Item is a function. */
					case FUNCTION:
						if (flag&ADDTREES){
							if (NOMEMORY==DsTermIndexing(ptr1,&ptr2->symbol->discr6[nkb])){
								return(NOMEMORY);
							}
						} else {
							if (NOMEMORY==DeleteBranch(ptr1,&ptr2->symbol->discr6[nkb])){
								return(NOMEMORY);
							}
						}
						if (((ptr1==ptr3)||(ptr1==ptr4))&&(ptr1[0]&SELECTED)
								&&(ii)&&(dd)){
							if (flag&ADDTREES){
								if (NOMEMORY==DsTermIndexing(ptr1,&ptr2->symbol->discr5[nkb])){
									return(NOMEMORY);
								}
							} else {
								if (NOMEMORY==DeleteBranch(ptr1,&ptr2->symbol->discr5[nkb])){
									return(NOMEMORY);
								}
							}
						}
						break;
				}
			}
		}
	}

	return(0);
} /* EditDiscrTrees */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Add item to a discrimination tree)  OCJ
 *
 *    This function adds an item to a discrimination tree.
 *    The item may be a function or a literal.
 *
 *    As usual for discrimination trees each literal or term is described
 *    by a full path from the top node of a discrimination tree down to a
 *    tree leave with a preorder traversal description of the literal or
 *    term, that is, a depth first, left to right inspection of all involved
 *    subterms.
 *
 *    The top node of each discrimination tree is the haschain structure
 *    for the corresponding literal or term. The corresponding pointer
 *    in this structure points to the first child node. The tree is then
 *    made of several chunks of memory, each chunk comprising the whole
 *    path from a given node to an end leave taking the first child node
 *    in each branch. The first chunk starts in the first child node of
 *    the top node in the the tree. Every additional term added to the
 *    tree creates an additional chunk or adds a new entry to a dsendnode.
 *
 *    Nodes that are children of the same parent node are forward linked.
 *    All children nodes of a given parent node are the start of memory chunks
 *    except the first child (but note that the first child of the top node
 *    is an exception to this rule, as it is also the start of a memory
 *    chunk).
 *
 *    There are the following kind of nodes:
 *    - Nodes for variable place holders: They consist of a byte marker
 *      with the VARIABLE value and a dsvarnode structure.
 *    - Nodes for functions: They consist of a byte marker with the
 *      FUNCTION value and a dsfncnode structure.
 *    - Leave nodes: They consist of a byte marker with the UNITEND
 *      value and a dsendnode structure.
 *    - Dummy nodes: They consist of a byte marker with the DUMMYNODE
 *      value and a pointer to next node with the same parent node.
 *      These nodes are produced when deleting tree data corresponding
 *      to an item in order to simplify the deletion algorithm.
 *
 *    Nodes of type dsvarnode and dsfncnode include the clsbrkts field
 *    to store the of brackets closed just before the term of the node,
 *    that is, between the node and the preceding node. This serves to
 *    traverse a term in the tree corresponding to a variable in the
 *    query term when searching for unifications and instances (see
 *    QueryUI() function). To calculate its value the pending set is
 *    used. It contains the number of arguments pending to be traversed
 *    for each function that is pending to be closed.
 *
 *
 *
 *  ARGUMENTS:
 *
 *    item: Pointer to item.
 *    topptr: Address of the pointer to first child node in the tree.
 *            If the pointer is NULL on entry then the tree is empty
 *            and the pointer is updated to the address of the first
 *            node in the tree that is created by this function.
 *
 *  RETURNS:return(NOMEMORY);
 *
 *    0 if successful, NOMEMORY if not enough memory.
 *
 *--------------------------------------------------------------*/
int32_t DsTermIndexing(uint8_t *item,uint8_t **topptr){
	auto uint8_t *ptr1,*ptr2,*ptr3;                 /* Auxiliary pointers */
	auto uint8_t *ptr4,*ptr5;                       /* Auxiliary pointers */
	auto uint8_t **ptr7;                            /* Auxiliary pointer */
	auto int32_t *ptr8;                             /* Auxiliary pointer */
	auto uint8_t *chunk;                            /* Pointer to new chunk of memory for item */
	auto int32_t size;                              /* Size in bytes of chunk buffer */
	auto int32_t used;                              /* Used bytes in chunk buffer */
	auto uint8_t **symbolset;                       /* Pointer to set of symbol pointers in chunk dsendnode */
	auto int32_t *pending;                          /* Pointer to set of pending arguments for each open term */
	auto int32_t pndsize;                           /* Number of elements allocated for pending set */
	auto int32_t ii,jj,kk,mm,nn,pp;                 /* Auxiliary */

	/* Allocate new chunk for the item. Size is overestimated */
	/* and chunk will be resized when needed. */
	ii=((symbol *)&item[1])->nextoff-((symbol *)&item[1])->offset;
	size=ii+sizeof(dsendnode);
	if (NULL==(chunk=mallocmn(size))){
		return(NOMEMORY);
	}
	used=0;
	if (NULL==(symbolset=mallocmn(TREEPTR_CHUNK_SIZE*sizeof(void *)))){
		MYFREE(chunk);
		return(NOMEMORY);
	}

	/* Allocate memory for pending arguments. The pending set contains */
	/* the number of active terms arguments not yet traversed. An active */
	/* term is a term for which not all its subterms have been traversed. */
	if (NULL==(pending=mallocmn(PNDARGS_CHUNK_SIZE*sizeof(*pending)))){
		MYFREE(chunk);
		MYFREE(symbolset);
		return(NOMEMORY);
	}
	pndsize=PNDARGS_CHUNK_SIZE;

	/* Loop through subterms of item. */
	/* ptr1 = subterm of item. */
	/* ptr2 = current node in tree. */
	/* ptr4 = top node in branch of tree to which ptr2 belongs. */
	/* topptr is set to address of pointer to be set to chunk. */
	/* jj = 1 if if traversed path matches the item, 0 otherwise. */
	/* kk = Number of active terms in pending set. */
	/* mm = Number of closed bracket just before the node term. */
	if (item[0]&EQUALITY){
		ptr1=&item[1+sizeof(symbol)];
	} else {
		ptr1=&item[1+SHORTSYMBOLSIZE];
	}
	ptr2=ptr4=*topptr;
	jj=1;
	kk=0;
	mm=0;
	while ((ptr1-item)<ii){

		/* Previous subterms matched the tree. */
		/* Loop through children nodes to check for a new match. */
		if (jj){

			/* Loop through nodes with the same parent node. */
			for (ptr3=ptr2,ptr7=topptr,jj=0;ptr3!=NULL;){

				/* Check for a match. */
				if ((((VARIABLE|FUNCTION)&ptr1[0])==ptr3[0])&&((VARIABLE&ptr1[0])
						||(((symbol *)&ptr1[1])->symbol==((dsfncnode *)&ptr3[1])->symbol))){
					jj=1;
					if (ptr3!=ptr2){
						ptr2=ptr4=ptr3;
					}
					break;
				}

				/* Prepare next iteration. */
				switch (ptr3[0]){
					case VARIABLE:
						ptr7=&((dsvarnode *)&ptr3[1])->next;
						break;
					case FUNCTION:
						ptr7=&((dsfncnode *)&ptr3[1])->next;
						break;
					case DUMMYNODE:
						ptr7=(uint8_t **)&ptr3[1];
						break;
				}
				ptr3=*ptr7;
			}

			/* If not match found the set topptr to address of pointer that */
			/* must be set to chunk. */
			if (jj==0){
				topptr=ptr7;
			}
		}

		/* The tree mismatch has been already detected. */
		if (jj==0){

			/* Allocate additional memory if necessary. */
			nn=used+2+(VARIABLE&ptr1[0]?vnodesize:fnodesize)+sizeof(dsendnode);
			if (nn>size){
				size=nn;
				if (NULL==(ptr5=reallocmn(chunk,size))){
					MYFREE(chunk);
					MYFREE(symbolset);
					MYFREE(pending);
					return(NOMEMORY);
				}
				chunk=ptr5;
			}

			/* Add new tree node to chunk. */
			if (VARIABLE&ptr1[0]){
				chunk[used]=VARIABLE;
				((dsvarnode *)&chunk[used+1])->next=NULL;
				((dsvarnode *)&chunk[used+1])->clsbrkts=mm;
				used+=(1+vnodesize);
			} else {
				chunk[used]=FUNCTION;
				((dsfncnode *)&chunk[used+1])->next=NULL;
				((dsfncnode *)&chunk[used+1])->symbol=((symbol *)&ptr1[1])->symbol;
				((dsfncnode *)&chunk[used+1])->clsbrkts=mm;
				used+=(1+fnodesize);
			}
		}

		/* Calculate the number of brackets closed after current term. */
		mm=0;
		pp=1;
		if ((VARIABLE&ptr1[0])||(0==((symbol *)&ptr1[1])->symbol->arity)){
			while ((kk>0)&&(pp==1)){
				pending[kk-1]--;
				if (pending[kk-1]<=0){
					mm++;
					kk--;
				} else {
					pp=0;
				}
			}
		}

		/* Update pending set. */
		if ((FUNCTION&ptr1[0])&&(((symbol *)&ptr1[1])->symbol->arity>0)){
			if (kk==pndsize){
				pndsize+=PNDARGS_CHUNK_SIZE;
				if (NULL==(ptr8=reallocmn(pending,pndsize*sizeof(*pending)))){
					MYFREE(chunk);
					MYFREE(symbolset);
					MYFREE(pending);
					return(NOMEMORY);
				}
				pending=ptr8;
			}
			pending[kk]=((symbol *)&ptr1[1])->symbol->arity;
			kk++;
		}

		/* Update term pointer in item and node pointer for next iteration. */
		ptr1+=(VARIABLE&ptr1[0]?3:1+SHORTSYMBOLSIZE);
		if (jj){
			ptr2+=(1+(VARIABLE&ptr2[0]?vnodesize:fnodesize));
		}
	}

	/* Free pending set. */
	MYFREE(pending);

	/* Chunk is not empty. */
	if (used>0){

		/* Resize chunk to the exact size required. */
		nn=used+1+sizeof(dsendnode);
		if (nn!=size){
			size=nn;
			if (NULL==(ptr5=reallocmn(chunk,size))){
				MYFREE(chunk);
				MYFREE(symbolset);
				return(NOMEMORY);
			}
			chunk=ptr5;
		}

		/* Add leave node to chunk. */
		chunk[used]=UNITEND;
		((dsendnode *)&chunk[used+1])->size=TREEPTR_CHUNK_SIZE;
		((dsendnode *)&chunk[used+1])->used=1;
		((dsendnode *)&chunk[used+1])->symbol=symbolset;
		((dsendnode *)&chunk[used+1])->symbol[0]=item;

		/* Add chunk to tree. */
		*topptr=chunk;

	/* Chunk is empty. */
	} else {

		/* Free empty chunk memory. */
		MYFREE(chunk);

		/* Tree for the item is empty. The item has no parameters. */
		/* Set a chunk with a leave node and add the chunk to the tree. */
		if (ptr4==NULL){
			if (NULL==(chunk=mallocmn(1+sizeof(dsendnode)))){
				MYFREE(symbolset);
				return(NOMEMORY);
			}
			chunk[0]=UNITEND;
			((dsendnode *)&chunk[1])->size=TREEPTR_CHUNK_SIZE;
			((dsendnode *)&chunk[1])->used=1;
			((dsendnode *)&chunk[used+1])->symbol=symbolset;
			((dsendnode *)&chunk[1])->symbol[0]=item;
			*topptr=chunk;

		/* Tree for the item is not empty. */
		} else {

			/* Free set of symbol pointers of dsendnode. */
			MYFREE(symbolset);

			/* Allocate more space for pointers in leave node if needed. */
			if (((dsendnode *)&ptr2[1])->size==((dsendnode *)&ptr2[1])->used){
				if (NULL==(symbolset=reallocmn(((dsendnode *)&ptr2[1])->symbol,
						(TREEPTR_CHUNK_SIZE+((dsendnode *)&ptr2[1])->size)*sizeof(void *)))){
					return(NOMEMORY);
				}
				((dsendnode *)&ptr2[1])->size+=TREEPTR_CHUNK_SIZE;
				((dsendnode *)&ptr2[1])->symbol=symbolset;
			}

			/* Set a leave node pointer to item. */
			((dsendnode *)&ptr2[1])->symbol[((dsendnode *)&ptr2[1])->used]=item;
			((dsendnode *)&ptr2[1])->used++;
		}
	}

	return(0);
} /* DsTermIndexing */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Delete item in a discrimination tree)  OCJ
 *
 *    This function deletes the discrimination tree data related to
 *    a specified item in a specified tree.
 *
 *    This function uses backtracking and assigns a stack element
 *    for each new branch accessed. The stack element points to the
 *    node that has the same parent node as the top node in the
 *    new traversed branch and has the next field pointing to
 *    that top node.
 *
 *
 *  ARGUMENTS:
 *
 *    item: Pointer to item.
 *    topptr: Address of the pointer to first child node in the tree.
 *
 *  RETURNS:
 *
 *    0 if OK or NOMEMORY.
 *
 *--------------------------------------------------------------*/
int32_t DeleteBranch(uint8_t *item,uint8_t **topptr){
	auto dscstack *stack;                           /* Stack for backtracking */
	auto uint8_t *ptr1,*ptr2,*ptr3;                 /* Auxiliary pointer */
	auto uint8_t *ptr4,*ptr5,*ptr6,*ptr7;           /* Auxiliary pointer */
	auto int32_t ii,jj,kk;                          /* Auxiliary */

	/* Return if tree is empty. */
	if (NULL==*topptr){
		return(0);
	}

	/* Allocate stack memory. */
	if (NULL==(stack=mallocmn(sizeof(dscstack)+(sizeof(stkelem)*(STACK_CHUNK_SIZE-1))))){
		return(NOMEMORY);
	}
	stack->size=STACK_CHUNK_SIZE;
	stack->used=0;

	/* Loop through subterms of item. */
	/* ptr1 = subterm of item. */
	/* ptr2 = current node in tree. */
	/* ptr4 = top node in branch of tree to which ptr2 belongs. */
	/* ptr5 = node which next field points to ptr2. */
	/*        This is the node that will be pushed into the stack. */
	/* topptr is set to address of pointer to be set to chunk. */
	/* ii = length of item with all its subterms. */
	/* jj = 1 if if traversed path matches the item, 0 otherwise. */
	/* kk = 1 if match was not with the first node of the same parent node. */
	if (item[0]&EQUALITY){
		ptr1=&item[1+sizeof(symbol)];
	} else {
		ptr1=&item[1+SHORTSYMBOLSIZE];
	}
	ptr2=ptr4=*topptr;
	ii=((symbol *)&item[1])->nextoff-((symbol *)&item[1])->offset;
	while ((ptr1-item)<ii){

		/* Loop through children nodes. */
		for (ptr3=ptr2,ptr5=NULL,jj=kk=0;ptr3!=NULL;){

			/* Check for a match. */
			if ((((VARIABLE|FUNCTION)&ptr1[0])==ptr3[0])&&((VARIABLE&ptr1[0])
					||(((symbol *)&ptr1[1])->symbol==((dsfncnode *)&ptr3[1])->symbol))){
				jj=1;
				if (ptr3!=ptr2){
					ptr2=ptr4=ptr3;
					kk=1;
				}
				break;
			}

			/* Prepare next iteration. */
			ptr5=ptr3;
			switch (ptr3[0]){
				case VARIABLE:
					ptr3=((dsvarnode *)&ptr3[1])->next;
					break;
				case FUNCTION:
					ptr3=((dsfncnode *)&ptr3[1])->next;
					break;
				case DUMMYNODE:
					ptr3=*((uint8_t **)&ptr3[1]);
					break;
			}
		}

		/* If match not found then report error in logic and return NOMEMORY. */
		if (jj==0){
			MYFREE(stack);
			printf("Problem in program logic deleting tree branch.\n");
			return(NOMEMORY);
		}

		/* Update the stack if we are in a new branch. */
		if (kk){
			if (stack->size<=stack->used){
				kk=stack->size+STACK_CHUNK_SIZE;
				if (NULL==(ptr6=reallocmn(stack,sizeof(dscstack)+(sizeof(stkelem)*kk)))){
					MYFREE(stack);
					return(NOMEMORY);
				}
				stack=(dscstack *)ptr6;
				stack->size=kk;
			}
			stack->element[stack->used].node=ptr5;
			stack->used++;
		}

		/* Update term pointer in item and node pointer for next iteration. */
		ptr1+=(VARIABLE&ptr1[0]?3:1+SHORTSYMBOLSIZE);
		ptr2+=(1+((*ptr2)==VARIABLE?vnodesize:fnodesize));
	}

	/* Now ptr2 points to an end node. Scan for an item match. */
	for (ii=0;ii<((dsendnode *)&ptr2[1])->used;ii++){

		/* There is an item match. */
		if (item==((dsendnode *)&ptr2[1])->symbol[ii]){

			/* More than one item linked by end node. Replace the */
			/* item entry with the last entry and adjust the used field. */
			if (((dsendnode *)&ptr2[1])->used>1){
				((dsendnode *)&ptr2[1])->symbol[ii]=((dsendnode *)&ptr2[1])->symbol[((dsendnode *)&ptr2[1])->used-1];
				((dsendnode *)&ptr2[1])->used--;

			/* Only current item is linked by end node. */
			} else {

				/* Loop through branches that potentially can be freed or reallocated, */
				/* ptr4 points to top node in branch, */
				/* ptr2 points to leave node in branch. */
				while (ptr4!=NULL){

					/* Loop through nodes in branch and set ptr5 to the last node */
					/* in branch whose next field is not NULL. If there are no nodes */
					/* in branch with next field not NULL then ptr5 is set to NULL. */
					for (ptr3=ptr4,ptr5=NULL;(ptr3[0]!=UNITEND)&&(ptr3[0]!=DUMMYNODE);
							ptr3+=(1+((*ptr3)==VARIABLE?vnodesize:fnodesize))){
						if (((ptr3[0]==VARIABLE)&&(((dsvarnode *)&ptr3[1])->next!=NULL))
								||((ptr3[0]==FUNCTION)&&(((dsfncnode *)&ptr3[1])->next!=NULL))){
							ptr5=ptr3;
						}
					}

					/* The top node is the last node with the next field not NULL */
					/* or there are no nodes with the next field not NULL. */
					/* Free branch and prepare next branch iteration. */
					if ((ptr5==ptr4)||(ptr5==NULL)){

						/* Unlink branch from tree. */
						if (ptr5==ptr4){
							if (VARIABLE==ptr4[0]){
								ptr7=((dsvarnode *)&ptr4[1])->next;
							} else {
								ptr7=((dsfncnode *)&ptr4[1])->next;
							}
						} else {
							ptr7=NULL;
						}
						if (stack->used>0){
							ptr6=stack->element[stack->used-1].node;
							switch (ptr6[0]){
								case VARIABLE:
									((dsvarnode *)&ptr6[1])->next=ptr7;
									break;
								case FUNCTION:
									((dsfncnode *)&ptr6[1])->next=ptr7;
									break;
								case DUMMYNODE:
									*((uint8_t **)&ptr6[1])=ptr7;
									break;
							}
						} else {
							*topptr=ptr7;
						}

						/* Free branch. */
						if (ptr2[0]==UNITEND){
							MYFREE(((dsendnode *)&ptr2[1])->symbol);
						}
						MYFREE(ptr4);

						/* Prepare branch iteration using stack data. */
						if (stack->used>0){
							ptr2=stack->element[stack->used-1].node;
							if ((ptr2[0]==DUMMYNODE)&&((*((uint8_t **)&ptr2[1]))==NULL)){
								if (stack->used>1){
									ptr4=stack->element[stack->used-2].node;
									switch (ptr4[0]){
										case VARIABLE:
											ptr4=((dsvarnode *)&ptr4[1])->next;
											break;
										case FUNCTION:
											ptr4=((dsfncnode *)&ptr4[1])->next;
											break;
										case DUMMYNODE:
											ptr4=*((uint8_t **)&ptr4[1]);
											break;
									}
								} else {
									ptr4=*topptr;
								}
								stack->used-=1;
							} else {
								ptr4=NULL;
							}
						} else {
							ptr4=NULL;
						}

					/* The last node with the next field not NULL is not the top node. */
					/* Reallocate branch to strictly needed length and link it to tree. */
					/* The last node is converted to a dummy node and no more  */
					/* branch iterations are done. */
					} else {

						/* If leave node is an end node then free memory allocated to it. */
						if (ptr2[0]==UNITEND){
							MYFREE(((dsendnode *)&ptr2[1])->symbol);
						}

						/* Reallocate branch. */
						if (VARIABLE==ptr5[0]){
							ptr6=((dsvarnode *)&ptr5[1])->next;
						} else {
							ptr6=((dsfncnode *)&ptr5[1])->next;
						}
						if (NULL==(ptr7=reallocmn(ptr4,(ptr5-ptr4)+1+sizeof(uint8_t *)))){
							MYFREE(stack);
							return(NOMEMORY);
						}

						/* Convert node to dummy node. */
						ptr5=ptr7+(ptr5-ptr4);
						ptr5[0]=DUMMYNODE;
						*((uint8_t **)&ptr5[1])=ptr6;

						/* Link branch to tree. */
						if (stack->used>0){
							ptr6=stack->element[stack->used-1].node;
							switch (ptr6[0]){
								case VARIABLE:
									((dsvarnode *)&ptr6[1])->next=ptr7;
									break;
								case FUNCTION:
									((dsfncnode *)&ptr6[1])->next=ptr7;
									break;
								case DUMMYNODE:
									*((uint8_t **)&ptr6[1])=ptr7;
									break;
							}
						} else {
							*topptr=ptr7;
						}

						/* Leave branch iterations. */
						break;
					}
				}
			}
			MYFREE(stack);
			return(0);
		}
	}

	/* Report error in logic and return. */
	MYFREE(stack);
	printf("Problem in program logic deleting tree branch.\n");
	return(NOMEMORY);
} /* DeleteBranch */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Free discrimination tree branch memory)  OCJ
 *
 *    This function frees the memory allocated for a discrimination tree
 *    branch.
 *
 *    The function first frees memory allocated to other branches pointed
 *    by nodes in the current branch by calling itself recursively.
 *
 *
 *  ARGUMENTS:
 *
 *    branch: Pointer to first node of the branch.
 *
 *  RETURNS:
 *
 *    None.
 *
 *--------------------------------------------------------------*/
void FreeDiscrBranch(uint8_t *branch){
	auto uint8_t *ptr1;                             /* Auxiliary pointer */

	/* Return if branch is not allocated. */
	if (branch==NULL){
		return;
	}

	/* Branch traversing loop. */
	for (ptr1=branch;((*ptr1)!=UNITEND)&&((*ptr1)!=DUMMYNODE);
			ptr1+=(1+((*ptr1)==VARIABLE?vnodesize:fnodesize))){
		if ((*ptr1)==FUNCTION){
			FreeDiscrBranch(((dsfncnode *)&ptr1[1])->next);
		} else {
			FreeDiscrBranch(((dsvarnode *)&ptr1[1])->next);
		}
	}

	/* Free branch memory and return. */
	if ((*ptr1)==DUMMYNODE){
		FreeDiscrBranch(*((uint8_t **)&ptr1[1]));
	} else {
		MYFREE(((dsendnode *)&ptr1[1])->symbol);
	}
	MYFREE(branch);
	return;
} /* FreeDiscrBranch */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Get next generalization candidate item)  OCJ
 *
 *    This function gets the next candidate item that is a generalization
 *    candidate of a given item in a given clause by searching the
 *    discrimination tree whose top node is passed as an argument.
 *
 *    This function uses backtracking, using one stack element for each net
 *    node traversed, that is, previous nodes with the same parent node as
 *    the one being traversed are not pushed in the stack. The data.pndbrkts
 *    field of stack elements is not used by this function.
 *
 *    To get the first candidate the pointer in pstack must be set to NULL
 *    or its used field set to 0 by the caller. To get subsequent candidates
 *    the value returned in pstack must be kept.
 *
 *    It is the caller responsibility to free memory allocated for the stack
 *    by this function.
 *
 *    This function is called recursively.
 *
 *
 *  ARGUMENTS:
 *
 *    subterm: Pointer to subterm of item corresponding to current call.
 *    pcand: Address of pointer to generalization candidate item.
 *    pstack: Address of pointer to a dscstack structure that allows backtracking
 *            in a discrimination tree.
 *    depth: Tree depth corresponding to current call. It must be zero for the
 *           calls that are not from QueryGeneric() itself.
 *    topnode: Pointer to top node of tree to be used for search.
 *             Only used for calls with depth=0.
 *
 *  RETURNS:
 *
 *    1 if a valid generalization candidate was found.
 *    0 if no valid generalization candidate was found.
 *    NOMEMORY if not enough memory.
 *
 *--------------------------------------------------------------*/
int32_t QueryGeneric(uint8_t *subterm,uint8_t **pcand,dscstack **pstack,
		int32_t depth,uint8_t *topnode){
	auto uint8_t *ptr1,*ptr3;                       /* Auxiliary pointer */
	auto dscstack *ptr2;                            /* Auxiliary pointer */
	auto int32_t ii;                                /* Auxiliary */

	/* Return failure if maximum recursive calls has been reached. */
	if (depth>MAXRECURSIVE){
		return(0);
	}

	/* Allocate stack memory. */
	if ((*pstack)==NULL){
		if (NULL==(*pstack=mallocmn(sizeof(dscstack)+(sizeof(stkelem)*(STACK_CHUNK_SIZE-1))))){
			return(NOMEMORY);
		}
		(*pstack)->size=STACK_CHUNK_SIZE;
		(*pstack)->used=0;
	}

	/* There are no more nodes in the stack at this depth. */
	if (depth>=(*pstack)->used){

		/* We are in the top tree node. Get pointer to top tree node */
		/* from function argument. */
		if (depth==0){
			ptr1=topnode;

		/* We are not in the top tree node. Get pointer to corresponding node */
		/* from the previous node in the stack. */
		} else {
			ptr1=(*pstack)->element[depth-1].node;
			ptr1+=(1+(ptr1[0]==VARIABLE?vnodesize:fnodesize));
		}

		/* Push node in the stack. */
		if ((*pstack)->size<=(*pstack)->used){
			ii=(*pstack)->size+STACK_CHUNK_SIZE;
			if (NULL==(ptr2=reallocmn(*pstack,sizeof(dscstack)+(sizeof(stkelem)*ii)))){
				return(NOMEMORY);
			}
			*pstack=ptr2;
			(*pstack)->size=ii;
		}
		(*pstack)->element[depth].node=ptr1;
		(*pstack)->used++;

		/* If this is an end node then return the first symbol in the set of symbols. */
		if ((ptr1!=NULL)&&(ptr1[0]==UNITEND)){
			(*pstack)->element[depth].data.index=0;
			*pcand=((dsendnode *)&ptr1[1])->symbol[0];
			return(1);
		}

	/* There is a stack element corresponding to this depth. */
	} else {

		/* End node. Return as appropriate. */
		ptr1=(*pstack)->element[depth].node;
		if (ptr1[0]==UNITEND){
			if (((dsendnode *)&ptr1[1])->used>(1+(*pstack)->element[depth].data.index)){
				(*pstack)->element[depth].data.index++;
				*pcand=((dsendnode *)&ptr1[1])->symbol[(*pstack)->element[depth].data.index];
				return(1);
			} else {
				return(0);
			}
		}

		/* Recursive call for next node. */
		switch (QueryGeneric(NextItem(subterm,ptr1[0]&VARIABLE?OVERSUBTERMS:IMMED),
				pcand,pstack,depth+1,topnode)){
			case NOMEMORY:
				return(NOMEMORY);
				break;
			case 1:
				return(1);
				break;
		}

		/* No match found in recursive call. Get a pointer to next node */
		/* that has the same parent as the current node. */
		ptr1=(ptr1[0]==VARIABLE?((dsvarnode *)&ptr1[1])->next:((dsfncnode *)&ptr1[1])->next);
	}

	/* If no more nodes to scan then return no match. */
	if (ptr1==NULL){
		return(0);
	}

	/* Scan for a match in nodes with the same parent node as current node. */
	ptr3=NULL; /* Just to avoid compiler warnings. */
	while (ptr1!=NULL){

		/* If there is a match then set ptr3 to appropriate subterm of item */
		/* for next recursive call. Otherwise set ptr3 to NULL. */
		switch (ptr1[0]){
			case VARIABLE:
				ptr3=NextItem(subterm,OVERSUBTERMS);
				break;
			case FUNCTION:
				if ((FUNCTION&subterm[0])
						&&(((symbol *)&subterm[1])->symbol==((dsfncnode *)&ptr1[1])->symbol)){
					ptr3=NextItem(subterm,IMMED);
				} else {
					ptr3=NULL;
				}
				break;
			case DUMMYNODE:
				ptr3=NULL;
				break;
		}

		/* There is a match. Set the stack to new scenario */
		/* and make a recursive call. */
		if (ptr3!=NULL){
			(*pstack)->element[depth].node=ptr1;
			(*pstack)->used=depth+1;
			switch (QueryGeneric(ptr3,pcand,pstack,depth+1,topnode)){
				case NOMEMORY:
					return(NOMEMORY);
					break;
				case 1:
					return(1);
					break;
			}
		}

		/* Prepare next iteration. */
		switch (ptr1[0]){
			case VARIABLE:
				ptr1=((dsvarnode *)&ptr1[1])->next;
				break;
			case FUNCTION:
				ptr1=((dsfncnode *)&ptr1[1])->next;
				break;
			case DUMMYNODE:
				ptr1=*((uint8_t **)&ptr1[1]);
				break;
		}
	}

	/* No match found. */
	return(0);
} /* QueryGeneric */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Get next unification or instance candidate item)  OCJ
 *
 *    This function gets the next candidate item that is an unification
 *    or instance candidate of a given item in a given clause by searching
 *    the discrimination tree whose top node is passed as an argument.
 *
 *    This function uses backtracking, using one stack element for each net
 *    node traversed, that is, previous nodes with the same parent node as
 *    the one being traversed are not pushed in the stack.
 *
 *    To get the first candidate the pointer in pstack must be set to NULL
 *    or its used field set to 0 by the caller. To get subsequent candidates
 *    the value returned in pstack must be kept. The data.pndbrkts field of
 *    stack elements is used to store the number of pending brackets to be
 *    closed to terminate the traversing of a tree term corresponding to a
 *    variable in the query term. This field is increased every new traversed
 *    tree subterm that is a function of arity one or bigger.
 *
 *    It is the caller responsibility to free memory allocated for the stack
 *    by this function.
 *
 *    This function is called recursively.
 *
 *
 *  ARGUMENTS:
 *
 *    subterm: Pointer to subterm of item corresponding to current call.
 *    pcand: Address of pointer to candidate item.
 *    pstack: Address of pointer to a dscstack structure that allows backtracking
 *            in a discrimination tree.
 *    depth: Tree depth corresponding to current call. It must be zero for the
 *           calls that are not from QueryUI() itself.
 *    topnode: Pointer to top node of tree to be used for search.
 *             Only used for calls with depth=0.
 *    type: UNIFICATION for searching of an item that unifies with the query term.
 *          INSTANCE for searching of an item that is an instance of the query term.
 *
 *  RETURNS:
 *
 *    1 if a valid candidate was found.
 *    0 if no valid candidate was found.
 *    NOMEMORY if not enough memory.
 *
 *--------------------------------------------------------------*/
int32_t QueryUI(uint8_t *subterm,uint8_t **pcand,dscstack **pstack,
		int32_t depth,uint8_t *topnode,int32_t type){
	auto uint8_t *ptr1,*ptr3;                       /* Auxiliary pointers */
	auto dscstack *ptr2;                            /* Auxiliary pointer */
	auto int32_t pndbrkts;                          /* Pending brackets to be closed */
	auto int32_t ii;                                /* Auxiliary */
	#ifdef DEBUGCODE
	auto char *auxstr;
	#endif

	/* Return failure if maximum recursive calls has been reached. */
	if (depth>MAXRECURSIVE){
		return(0);
	}

	/* Allocate stack memory. */
	if ((*pstack)==NULL){
		if (NULL==(*pstack=mallocmn(sizeof(dscstack)+(sizeof(stkelem)*(STACK_CHUNK_SIZE-1))))){
			return(NOMEMORY);
		}
		(*pstack)->size=STACK_CHUNK_SIZE;
		(*pstack)->used=0;
	}

	/* There are no more nodes in the stack at this depth. */
	ii=0; /* Just to avoid compiler warnings. */
	if (depth>=(*pstack)->used){

		/* We are in the top tree node. Get pointer to top tree node */
		/* from function argument. */
		if (depth==0){
			ptr1=topnode;

		/* We are not in the top tree node. Get pointer to corresponding node */
		/* from the previous node in the stack. */
		} else {
			ptr1=(*pstack)->element[depth-1].node;
			ptr1+=(1+(ptr1[0]==VARIABLE?vnodesize:fnodesize));
		}

		/* We have a tree node. */
		if (ptr1!=NULL){

			/* Push node in the stack except for the pndbrkts and subterm fields. */
			if ((*pstack)->size<=(*pstack)->used){
				ii=(*pstack)->size+STACK_CHUNK_SIZE;
				if (NULL==(ptr2=reallocmn(*pstack,sizeof(dscstack)+(sizeof(stkelem)*ii)))){
					return(NOMEMORY);
				}
				*pstack=ptr2;
				(*pstack)->size=ii;
			}
			(*pstack)->element[depth].node=ptr1;
			(*pstack)->used++;

			/* If this is an end node then return the first symbol in the set of symbols. */
			if (ptr1[0]==UNITEND){
				(*pstack)->element[depth].data.index=0;
				*pcand=((dsendnode *)&ptr1[1])->symbol[0];
				return(1);
			}

			/* Calculate number of pending brackets that remain to be closed */
			/* just before the term of the current node after applying */
			/* the applicable closing brackets. */
			/* WARNING-WARNING-WARNING: If this code is modified it must */
			/* be modified also in the appropriate place below. */
			if ((depth==0)||(ptr1[0]==UNITEND)){
				pndbrkts=0;
			} else {
				pndbrkts=(*pstack)->element[depth-1].data.pndbrkts;
				switch (ptr1[0]){
					case DUMMYNODE:
						ptr3=*((uint8_t **)&ptr1[1]);
						ii=(ptr3[0]==VARIABLE?((dsvarnode *)&ptr3[1])->clsbrkts:
								((dsfncnode *)&ptr3[1])->clsbrkts);
						break;
					case VARIABLE:
						ii=((dsvarnode *)&ptr1[1])->clsbrkts;
						break;
					case FUNCTION:
						ii=((dsfncnode *)&ptr1[1])->clsbrkts;
						break;
				}
				if (ii>=pndbrkts){
					pndbrkts=0;
				} else {
					pndbrkts-=ii;
				}
			}

			/* Update subterm argument to the effective value for this node and save it */
			/* in the stack. This value is not correct if in the previous node we were */
			/* traversing node terms for a variable in the query term but in this node the */
			/* traversing is finished. */
			if ((pndbrkts==0)&&(depth>0)&&((*pstack)->element[depth-1].data.pndbrkts>0)){
				subterm=NextItem(subterm,OVERSUBTERMS);
			}
			(*pstack)->element[depth].subterm=subterm;
		}

	/* There is a stack element corresponding to this depth. */
	} else {

		/* End node. Return as appropriate. */
		ptr1=(*pstack)->element[depth].node;
		if (ptr1[0]==UNITEND){
			if (((dsendnode *)&ptr1[1])->used>(1+(*pstack)->element[depth].data.index)){
				(*pstack)->element[depth].data.index++;
				*pcand=((dsendnode *)&ptr1[1])->symbol[(*pstack)->element[depth].data.index];
				return(1);
			} else {
				return(0);
			}
		}

		/* Debug: Get information about the tree node. */
		#ifdef DEBUGCODE
		switch (ptr1[0]){
			case VARIABLE:
				auxstr=NULL;
				break;
			case UNITEND:
				auxstr=NULL;
				break;
			case FUNCTION:
				auxstr=((char *)((dsfncnode *)&ptr1[1])->symbol)+(sizeof(hashchain)+(SPACE_AFTER_HASHCHAIN*sizeof(void *)));
				break;
		}
		ptr3=(uint8_t *)auxstr; /* To prevent compiler warnings. */
		#endif

		/* Update subterm argument with the value in the stack. This is necessary */
		/* as some updates to effective subterm cannot be done while traversing */
		/* the stack. */
		subterm=(*pstack)->element[depth].subterm;

		/* Recursive call for next node. */
		if ((*pstack)->element[depth].data.pndbrkts>0){
			ptr3=subterm;
		} else {
			if (VARIABLE&ptr1[0]){
				ptr3=NextItem(subterm,OVERSUBTERMS);
			} else {
				ptr3=NextItem(subterm,IMMED);
			}
		}
		switch (QueryUI(ptr3,pcand,pstack,depth+1,topnode,type)){
			case NOMEMORY:
				return(NOMEMORY);
				break;
			case 1:
				return(1);
				break;
		}

		/* No match found in recursive call. Get a pointer to next node */
		/* that has the same parent as the current node. */
		ptr1=(ptr1[0]==VARIABLE?((dsvarnode *)&ptr1[1])->next:((dsfncnode *)&ptr1[1])->next);

		/* Calculate number of pending brackets that remain to be closed */
		/* just before the term of the current node after applying */
		/* the applicable closing brackets. */
		/* WARNING-WARNING-WARNING: If this code is modified it must */
		/* be modified also in the appropriate place above. */
		if (ptr1!=NULL){
			if ((depth==0)||(ptr1[0]==UNITEND)){
				pndbrkts=0;
			} else {
				pndbrkts=(*pstack)->element[depth-1].data.pndbrkts;
				switch (ptr1[0]){
					case DUMMYNODE:
						ptr3=*((uint8_t **)&ptr1[1]);
						ii=(ptr3[0]==VARIABLE?((dsvarnode *)&ptr3[1])->clsbrkts:
								((dsfncnode *)&ptr3[1])->clsbrkts);
						break;
					case VARIABLE:
						ii=((dsvarnode *)&ptr1[1])->clsbrkts;
						break;
					case FUNCTION:
						ii=((dsfncnode *)&ptr1[1])->clsbrkts;
						break;
				}
				if (ii>=pndbrkts){
					pndbrkts=0;
				} else {
					pndbrkts-=ii;
				}
			}
		}
	}

	/* Now ptr1 points to current node that is not in the stack */
	/* or is NULL if there are no more nodes to scan. */
	/* If no more nodes to scan then return no match. */
	if (ptr1==NULL){
		return(0);
	}

	/* If we are traversing a tree term that is matching a variable in the */
	/* query term then remember it to corresponding term in the tree. */
	if (pndbrkts>0){
		ii=0;
		ptr3=subterm;

	/* Otherwise remember it. */
	} else {
		ii=1;
		ptr3=NULL; /* Just to avoid compiler warnings. */
	}

	/* Scan for a match in nodes with the same parent node as current node. */
	while (ptr1!=NULL){

		/* Debug: Get information about the tree node. */
		#ifdef DEBUGCODE
		if (VARIABLE==ptr1[0]){
			auxstr=NULL;
		} else if (FUNCTION==ptr1[0]) {
			auxstr=((char *)((dsfncnode *)&ptr1[1])->symbol)+(sizeof(hashchain)+(SPACE_AFTER_HASHCHAIN*sizeof(void *)));
		} else {
			auxstr=NULL;
		}
		#endif

		/* If we are not traversing a tree term that is matching a variable */
		/* in the query term then check for a new match and if so then set ptr3 */
		/* to appropriate subterm of item for next recursive call. Otherwise */
		/* set ptr3 to NULL. */
		if (ii){
			switch (ptr1[0]){
				case VARIABLE:
					if ((type==UNIFICATION)||(VARIABLE&subterm[0])){
						ptr3=NextItem(subterm,OVERSUBTERMS);
					} else {
						ptr3=NULL;
					}
					break;
				case FUNCTION:
					if (VARIABLE&subterm[0]){
						ptr3=subterm;
					} else if ((FUNCTION&subterm[0])
							&&(((symbol *)&subterm[1])->symbol==((dsfncnode *)&ptr1[1])->symbol)){
						ptr3=NextItem(subterm,IMMED);
					} else {
						ptr3=NULL;
					}
					break;
				case DUMMYNODE:
					ptr3=NULL;
					break;
			}
		}

		/* There is a match. */
		if ((ptr3!=NULL)&&(ptr1[0]!=DUMMYNODE)){

			/* Update stack to new scenario. */
			(*pstack)->element[depth].node=ptr1;
			(*pstack)->used=depth+1;
			if (ii){
				if ((VARIABLE&subterm[0])&&(ptr1[0]==FUNCTION)
						&&(((dsfncnode *)&ptr1[1])->symbol->arity>0)){
					(*pstack)->element[depth].data.pndbrkts=1;
				} else {
					(*pstack)->element[depth].data.pndbrkts=0;
					if (VARIABLE&subterm[0]){
						ptr3=NextItem(subterm,IMMED);
					}
				}
			} else {
				if ((ptr1[0]==FUNCTION)&&(((dsfncnode *)&ptr1[1])->symbol->arity>0)){
					(*pstack)->element[depth].data.pndbrkts=pndbrkts+1;
				} else {
					(*pstack)->element[depth].data.pndbrkts=pndbrkts;
				}
			}

			/* Recursive call. */
			switch (QueryUI(ptr3,pcand,pstack,depth+1,topnode,type)){
				case NOMEMORY:
					return(NOMEMORY);
					break;
				case 1:
					return(1);
					break;
			}
		}

		/* Prepare next iteration. */
		switch (ptr1[0]){
			case VARIABLE:
				ptr1=((dsvarnode *)&ptr1[1])->next;
				break;
			case FUNCTION:
				ptr1=((dsfncnode *)&ptr1[1])->next;
				break;
			case DUMMYNODE:
				ptr1=*((uint8_t **)&ptr1[1]);
				break;
		}
	}

	/* No match found. */
	return(0);
} /* QueryUI */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Unlink binary clause from KB)  OCJ
 *
 *    This function unlinks a clause in binary form from all
 *    queues of the given KB, with the exception of the disposal
 *    queue. This function doesn't remove the clause from the
 *    disposal queue. In fact it is assumed that the clause is
 *    not linked to disposal queue.
 *
 *    If the clause is ACTIVE or PASSIVE then it is also removed
 *    from discrimination trees.
 *
 *    IMPORTANT: This function must never be called for main KB.
 *    if this changes in the future then the set of pkb variable
 *    must be changed accordingly.
 *
 *
 *  ARGUMENTS:
 *
 *    clause: Pointer to clause in binary form.
 *    nkb: Knowledge to which the clause belongs.
 *
 *  RETURNS:
 *
 *    0 if successful, NOMEMORY if not enough memory.
 *
 *--------------------------------------------------------------*/
int32_t Unlink(cmprefix *clause,int32_t nkb){
	auto kbase *pkb;                                     /* Pointer to KB */
	auto int32_t ii,jj,kk;                               /* Auxiliary */

	/* Remove clause from discrimination trees if it is ACTIVE or */
	/* it is PASSIVE and clause is in main KB or in a KB running */
	/* Otter algorithm. */
	if ((clause->flags&ACTIVE)||(((clause->flags&PASSIVE))
			&&((nkb==num_wkbs)||(kbset[nkb]->opts.algorithm==OTTER)))){
		if (NOMEMORY==EditDiscrTrees(clause,nkb,DELETETREES)){
			return(NOMEMORY);
		}
	}

	/* Unchain clause assertions. */
	UnchainAsserts(clause->part2.bin->asserts);

	/* Get address of pointer to KB. */
	if (nkb==num_wkbs){
		pkb=mainkb;
	} else {
		pkb=kbset[nkb];
	}

	/* Unlink formula. */
	switch (clause->flags&(ACTIVE|PASSIVE|UNPROC|LOCKED)){

		/* Unprocessed clause. Unlink clause from UNPROC queue. */
		case UNPROC:
			if (pkb->lstunproc==clause){
				pkb->lstunproc=clause->prev;
			}
			if (pkb->frstunproc==clause){
				pkb->frstunproc=clause->next;
			}
			pkb->prstats.nbunproc--;
			break;

		/* Locked clause. Unlink clause from LOCKED queue. */
		case LOCKED:
			if (pkb->lstlocked==clause){
				pkb->lstlocked=clause->prev;
			}
			if (pkb->frstlocked==clause){
				pkb->frstlocked=clause->next;
			}
			break;

		/* ACTIVE clause. Unlink clause from active queue */
		/* and update number of active clauses in KB. */
		case ACTIVE:
			if (pkb->lstactive==clause){
				pkb->lstactive=clause->prev;
			}
			if (pkb->frstactive==clause){
				pkb->frstactive=clause->next;
			}
			pkb->prstats.nbactive--;
			break;

		/* PASSIVE clause. */
		case PASSIVE:

			/* Unlink clause from passive queue. */
			if (pkb->lstpassive==clause){
				pkb->lstpassive=clause->prev;
			}
			if (pkb->frstpassive==clause){
				pkb->frstpassive=clause->next;
			}

			/* Unlink clause from weight queue. */
			for (kk=(NULL==kbset[nkb]->lrnvector?1:0);kk<WEIGHTQUEUES;kk++){
				ii=(clause->part2.bin->clweight[kk]>MAXWEIGHT?MAXWEIGHT:clause->part2.bin->clweight[kk]);
				if (pkb->frstweight[kk][ii]==clause){
					if (clause->part2.bin->nextweight[kk]!=NULL){
						jj=(clause->part2.bin->nextweight[kk]->part2.bin->clweight[kk]>MAXWEIGHT?
								MAXWEIGHT:clause->part2.bin->nextweight[kk]->part2.bin->clweight[kk]);
						if (ii==jj){
							pkb->frstweight[kk][ii]=clause->part2.bin->nextweight[kk];
						} else {
							pkb->frstweight[kk][ii]=NULL;
						}
					} else {
						pkb->frstweight[kk][ii]=NULL;
					}
				}
				if (pkb->lstweight[kk][ii]==clause){
					if (clause->part2.bin->prevweight[kk]!=NULL){
						jj=(clause->part2.bin->prevweight[kk]->part2.bin->clweight[kk]>MAXWEIGHT?
								MAXWEIGHT:clause->part2.bin->prevweight[kk]->part2.bin->clweight[kk]);
						if (ii==jj){
							pkb->lstweight[kk][ii]=clause->part2.bin->prevweight[kk];
						} else {
							pkb->lstweight[kk][ii]=NULL;
						}
					} else {
						pkb->lstweight[kk][ii]=NULL;
					}
				}
				if (clause->part2.bin->nextweight[kk]!=NULL){
					clause->part2.bin->nextweight[kk]->part2.bin->prevweight[kk]=clause->part2.bin->prevweight[kk];
				}
				if (clause->part2.bin->prevweight[kk]!=NULL){
					clause->part2.bin->prevweight[kk]->part2.bin->nextweight[kk]=clause->part2.bin->nextweight[kk];
				} else {
					pkb->frstglweight[kk]=clause->part2.bin->nextweight[kk];
				}
			}

			/* Update number of passive clauses in KB. */
			pkb->prstats.nbpassive--;
			break;
	}

	/* Common unlink tasks. */
	if (clause->prev!=NULL){
		clause->prev->next=clause->next;
	}
	if (clause->next!=NULL){
		clause->next->prev=clause->prev;
	}

	return(0);
} /* Unlink */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Free KB tree memory)  OCJ
 *
 *    This function frees all memory allocated for the trees of a
 *    specific KB.
 *
 *
 *  ARGUMENTS:
 *
 *    nkb: KB number.
 *
 *  RETURNS:
 *
 *    None.
 *
 *--------------------------------------------------------------*/
void FreeTreeMemory(int32_t nkb){
	auto hashchain *ptr3;                           /* Auxiliary pointer */

	/* Free tree memory. */
	for (ptr3=kb_hash;ptr3!=NULL;ptr3=ptr3->nextelem){
		FreeDiscrBranch(ptr3->discr1[nkb]);
		FreeDiscrBranch(ptr3->discr2[nkb]);
		FreeDiscrBranch(ptr3->discr3[nkb]);
		FreeDiscrBranch(ptr3->discr4[nkb]);
		FreeDiscrBranch(ptr3->discr5[nkb]);
		FreeDiscrBranch(ptr3->discr6[nkb]);
		ptr3->discr1[nkb]=ptr3->discr2[nkb]=ptr3->discr3[nkb]=ptr3->discr4[nkb]=
				ptr3->discr5[nkb]=ptr3->discr6[nkb]=NULL;
	}
	if (equptrs!=NULL){
		FreeDiscrBranch(equkey.discr1[nkb]);
		FreeDiscrBranch(equkey.discr2[nkb]);
		FreeDiscrBranch(equkey.discr3[nkb]);
		FreeDiscrBranch(equkey.discr4[nkb]);
		FreeDiscrBranch(equkey.discr5[nkb]);
		FreeDiscrBranch(equkey.discr6[nkb]);
		equkey.discr1[nkb]=equkey.discr2[nkb]=equkey.discr3[nkb]=equkey.discr4[nkb]=
				equkey.discr5[nkb]=equkey.discr6[nkb]=NULL;
	}

	return;
} /* FreeTreeMemory */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Calculate 32 bits hash value for a clause key)  OCJ
 *
 *    This function calculates a couple of hash values for a clause
 *    formula.
 *
 *    The hash value is guaranteed to be the same for variants of the
 *    same clause as defined in GetSplitSymbol() function.
 *
 *
 *  ARGUMENTS:
 *
 *    formula: Pointer to component binary formula.
 *
 *  RETURNS:
 *
 *    32 bits hash value.
 *
 *--------------------------------------------------------------*/
uint32_t HashClause(uint8_t *formula){
	auto uint32_t hash;                                  /* 32 bits hash value for the key clause */
	auto uint8_t *ptr1,*ptr2,*ptr3;                      /* Auxiliary pointers */
	auto uint32_t ii;                                    /* Auxiliary */
	auto int32_t jj;                                     /* Auxiliary */

	/* Generate hash values for each literal in clause. */
	/* These values are added together to generate the */
	/* clause hash values. */
	/* Loop through literals. */
	hash=0;
	for (ptr1=formula;(*ptr1)!=UNITEND;ptr1=NextItem(ptr1,OVERSUBTERMS)){

		/* Literal is an equality or inequality. */
		if (ptr1[0]&EQUALITY){

			/* Add the equality hash value for the equality itself. */
			ii=((symbol *)&ptr1[1])->symbol->hash;
			if (ptr1[0]&NEGATED){
				ii^=0xe6546b64;
			}
			hash+=(ii^hashpos[0]);

			/* Calculate hash value for first equality root term. This hash value is */
			/* calculated by xoring the hash value of each item with the hash value */
			/* corresponding to the item position in the root term and adding the results. */
			/* Loop through first equality root term items. */
			for (ptr3=NextItem(ptr2=NextItem(ptr1,IMMED),OVERSUBTERMS),jj=0;
					ptr2<ptr3;ptr2=NextItem(ptr2,IMMED),jj++){
				if (ptr2[0]&FUNCTION){
					ii=((symbol *)&ptr2[1])->symbol->hash;
				} else {
					ii=0xcc9e2d51;
				}
				hash+=(ii^hashpos[jj<MAXPOSITIONS?jj:MAXPOSITIONS-1]);
			}

			/* Calculate hash value for second equality root term. This hash value is */
			/* calculated by xoring the hash value of each item with the hash value */
			/* corresponding to the item position in the root term and adding the results. */
			/* Loop through first equality root term items. */
			for (ptr3=NextItem(ptr2=ptr3,OVERSUBTERMS),jj=0;
					ptr2<ptr3;ptr2=NextItem(ptr2,IMMED),jj++){
				if (ptr2[0]&FUNCTION){
					ii=((symbol *)&ptr2[1])->symbol->hash;
				} else {
					ii=0xcc9e2d51;
				}
				hash+=(ii^hashpos[jj<MAXPOSITIONS?jj:MAXPOSITIONS-1]);
			}

		/* Literal is a predicate. */
		/* Calculate hash value for literal. This hash value is calculated */
		/* by xoring the hash value of each item with the hash value corresponding */
		/* to the item position in the literal and adding the results. */
		/* Loop through literal items. */
		} else {
			for (ptr2=ptr1,ptr3=NextItem(ptr1,OVERSUBTERMS),jj=0;ptr2<ptr3;ptr2=NextItem(ptr2,IMMED),jj++){
				if (ptr2[0]&(PREDICATE|EQUALITY|FUNCTION)){
					ii=((symbol *)&ptr2[1])->symbol->hash;
				} else {
					ii=0xcc9e2d51;
				}
				if (ptr2[0]&NEGATED){
					ii^=0xe6546b64;
				}
				hash+=(ii^hashpos[jj<MAXPOSITIONS?jj:MAXPOSITIONS-1]);
			}
		}
	}

	return(hash);
} /* HashClause */
