/*
 ============================================================================
 Name        : multith.c
 Author      : Oscar Contreras
 Version     :
 Copyright   : Copyright (C) 2015-2016 Oscar Contreras. All rights reserved.
 Description : First order logic theorem prover in C
 ============================================================================
 */

/****************************************************************
*
*               multith (Multithreading functions module)
*
* All of the documentation and software included in the Drodi package
* is copyrighted by Oscar Contreras.
*
* Copyright (C) 2015-2016 Oscar Contreras. All rights reserved.
*
*
* This software is provided "as-is," without any express or implied warranty.
* In no event shall the author be held liable for any damages arising from the
* use, or the result of use, of this software and its documentation in terms
* of correctness, accuracy, reliability, currentness, or otherwise.
*
* Permission is granted to anyone to use this software for any purpose,
* including commercial use, and to alter and redistribute it, provided
* that the following conditions are met:
*
* 1. All redistributions of source code files must retain all copyright
*    notices that are currently in place, and this list of conditions without
*    modification.
*
* 2. All redistributions in binary form must retain all occurrences of the
*    above copyright notice that are currently in place (for example, in
*    the About boxes).
*
* 3. The origin of this software must not be misrepresented; you must not
*    claim that you wrote the original software.
*
* 4. Modified versions in source or binary form must be plainly marked as
*    such, and must not be misrepresented as being the original software.
*
*
* Oscar Contreras
* rwtrainer10 AT gmail.com
*
****************************************************************/

/*-------------------------------------------------------------
 *
 *  DESCRIPTION:
 *
 *    This source contains the multithreading functions, saturation algorithms
 *    and other related functions of the Drodi package.
 *
 *    Drodi is a multi-platform Open Source Artificial Intelligence application.
 *    See Drodi.c for additional information.
 *
 *
 *  IMPLEMENTATION:
 *
 *    Multi-platform
 *
 *  NOTES:
 *
 *
 *--------------------------------------------------------------*/

/* Include for all C sources */
#include "global.h"

/** Global variables for this module: only those that need initialization in their definition. */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Start saturation algorithms)  OCJ
 *
 *    This function perform the following tasks:
 *    - Converts the input formulas to conjunctive normal form, compiles
 *      the produced clauses and adds them to the main KB.
 *    - Makes copies of the resulting KB.
 *    - Start a saturation algorithm with each KB. Each algorithm is run
 *      in a separate thread and there are as many threads as available
 *      cores.
 *    - Prints the resulting output.
 *
 *
 *  ARGUMENTS:
 *
 *    None.
 *
 *  RETURNS:
 *
 *    None.
 *
 *--------------------------------------------------------------*/
void Saturate(void){
	auto int32_t ii;                                 /* Auxiliary */

	/* Start time measurement. */
	procctl.start=clock();
	gettimeofday(&mainkb->time,NULL);

	/* No KB's available. */
	if (mainkb==NULL){
		if (pgmmode==NONINTERACTIVE){
			printf("%% SZS status MemoryOut : No memory available\n");
		} else {
			printf("Cannot process: No KB's available\n");
		}
		return;
	}

	/* Main KB is empty. */
	if (mainkb->nformulas==0){
		if (pgmmode==NONINTERACTIVE){
			printf("%% SZS status InputError : ");
		}
		printf("There are no input formulas to process.\n");
		return;
	}

	/* Perform pre-processing. */
	ii=Preprocess(1);
	if (ii){
		if (ii==NOMEMORY){
			if (pgmmode==NONINTERACTIVE){
				printf("%% SZS status MemoryOut : ");
			}
			printf("Not enough memory.\n");
		} else {
			if (pgmmode==NONINTERACTIVE){
				printf("%% SZS status Timeout : ");
			}
			printf("Process timeout\n");
		}
		FreeBinFormulas(num_wkbs);
		if ((NOMEMORY==Initialize_KB(num_wkbs))&&(pgmmode==INTERACTIVE)){
			printf("No memory available, KB not initialized\n");
		}
		ResetStats(&mainkb->prstats);
		procctl.status=STOPPED;
		tot_pbms++;
		procctl.end=clock();
		tot_cpu_time+=((double)(procctl.end-procctl.start))/CLOCKS_PER_SEC;
		return;
	}

	/* Start time measurement after preprocessing. */
	gettimeofday(&mainkb->time,NULL);

	/* Select strategies. */
	if (0!=InitStrategies()){
		if (pgmmode==NONINTERACTIVE){
			printf("%% SZS status MemoryOut : Not enough memory available\n");
		} else {
			printf("Cannot process: Not enough memory available\n");
		}
		return;
	}

	/* Initialize locks. */
	omp_init_lock(&procctl.omp_lockc1);
	omp_init_lock(&procctl.omp_lockc2);

	/* Set as many threads as number of enabled cores and start threads. */
	omp_set_num_threads(num_wkbs);
	procctl.status=RUNNING;
	#pragma omp parallel default(shared)
	{
		InferenceThread();
	}

	/* Destroy locks. */
	omp_destroy_lock(&procctl.omp_lockc1);
	omp_destroy_lock(&procctl.omp_lockc2);

	/* Accumulate saturation processing elapsed time. */
	gettimeofday(&mainkb->endtime,NULL);
	mainkb->prstats.elapsed_time+=(mainkb->endtime.tv_sec-mainkb->time.tv_sec
			+(mainkb->endtime.tv_usec-mainkb->time.tv_usec)/1000000.0);

	/* If a non critical interruption is in process adjust the signalflag2 */
	/* value to indicate that saturation is finished. */
	if (signalflag2){
		signalflag2+=6;
	}

	/* Stop time measurement. */
	procctl.end=clock();
	cpu_time+=((double)(procctl.end-procctl.start))/CLOCKS_PER_SEC;
	tot_cpu_time+=cpu_time;

	/* Display KB data.*/
	if (prtkbdata!=PRTOFF){
		printf("\nKB DATA DUMP\n");
		printf("============\n");
		printf("Process status: ");
		switch (procctl.status){
			case STOPPED:
				printf("Stopped with no results obtained\n");
				break;
			case SATISFIABLE:
				printf("Stopped and satisfiable\n");
				break;
			case UNSATISFIABLE:
				printf("Stopped and unsatisfiable\n");
				break;
			case UNKNOWN:
				printf("Stopped and problem is unknown\n");
				break;
			case TIMEOUT:
				printf("Stopped because a timeout condition\n");
				break;
			case NOMEMORY:
				printf("Stopped because a no memory condition\n");
				break;
		}
		if (prtkbdata==PRTWINNER){
			if (procctl.solvekb>=0){
				if (((procctl.status!=NOMEMORY)&&(procctl.status!=TIMEOUT))||(DEBUGKBDATA!=0)){
					printf("Winning thread and KB: %d\n",procctl.solvekb);
					PrintKBData(procctl.solvekb);
				} else {
					printf("Remaining KB data not printed because abnormal end.\n");
				}
			}
		} else {
			if (((procctl.status!=NOMEMORY)&&(procctl.status!=TIMEOUT))||(DEBUGKBDATA!=0)){
				for (ii=0;ii<num_wkbs;ii++){
					printf("KB %d data:\n",ii);
					PrintKBData(ii);
				}
			} else {
				printf("Remaining KB data not printed because abnormal end.\n");
			}
		}
	}

	/* Display statistics. The PrintStats() function initializes the working KB's. */
	/* Initialize CPU time. */
	PrintStats(0);
	cpu_time=0.0;

	/* Accumulate global statistics. After the PrintStats(0) call */
	/* the main KB has the current problem accumulated statistics. */
	tot_pbms++;
	if ((procctl.status==SATISFIABLE)||(procctl.status==UNSATISFIABLE)){
		solved_pbms++;
	}
	AddInferenceStats(&mainkb->prstats,&glblstats);
	glblstats.elapsed_time+=mainkb->prstats.elapsed_time;
	glblstats.numitems++;
	if (glblstats.memory<mainkb->prstats.memory){
		glblstats.memory=mainkb->prstats.memory;
	}

	/* Reset working KB's statistics. */
	for (ii=0;ii<num_wkbs;ii++){
		ResetStats(&kbset[ii]->prstats);
	}

	/* Memory problems DEBUG. */
	#ifdef DEBUGMEMORY
	for (ii=0;ii<ndemods;ii++){
		if (memtrack[ii]==1){
			printf("Memory allocation number %d not freed.\n",ii);
		}
	}
	printf("Number of memory allocation calls: %d\n",ndemods);
	#endif

	/* Initialize main KB. */
	FreeBinFormulas(num_wkbs);
	if (NOMEMORY==Initialize_KB(num_wkbs)){
		printf("No memory available, KB not initialized\n");
	}
	ResetStats(&mainkb->prstats);
	procctl.status=STOPPED;

	return;
} /* Saturate */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Set strategy specific KB parameters)  OCJ
 *
 *    This function sets strategy specific parameters for a given
 *    strategy. See InitStrategies() general comments for details
 *    on how strategies are selected for each problem type.
 *
 *
 *  ARGUMENTS:
 *
 *    strategyidx: Index to strategies[] set.
 *    nkb: KB for which the strategy parameters will be set.
 *
 *  RETURNS:
 *
 *    None.
 *
 *
 *--------------------------------------------------------------*/
void SetMTParams(int32_t strategyidx,int32_t nkb){
	auto int32_t mm;                                     /* Auxiliary */
	auto uint16_t kk;                                    /* Auxiliary */

	/* Set algorithm. */
	kk=strategies[strategyidx];
	if (OTTER_P==(ALGMASK&kk)){
		kbset[nkb]->opts.algorithm=OTTER;
	} else {
		kbset[nkb]->opts.algorithm=DISCOUNT;
	}

	/* Set literal selection options. */
	if (SINGLENEG_P==(SELECTMASK&kk)){
		kbset[nkb]->opts.select=SINGLENEG;
	} else if (MULTINEG_P==(SELECTMASK&kk)){
		kbset[nkb]->opts.select=MULTINEG;
	} else if (SINGLEPOS_P==(SELECTMASK&kk)){
		kbset[nkb]->opts.select=SINGLEPOS;
	} else if (MULTIPOS_P==(SELECTMASK&kk)){
		kbset[nkb]->opts.select=MULTIPOS;
	} else {
		kbset[nkb]->opts.select=MAXIMAL;
	}

	/* Set weight factors option for clause weight calculation. */
	if (WFACT1_P==(WFACTSETMASK&kk)){
		kbset[nkb]->opts.factoridx=1;
	} else {
		kbset[nkb]->opts.factoridx=2;
	}

	/* Set equivalent factoring inference enabled/disabled. */
	if (EQFACTON_P==(EQFACTMASK&kk)){
		kbset[nkb]->opts.equivfact=1;
	} else {
		kbset[nkb]->opts.equivfact=0;
	}

	/* Set factoring inferences enabled/disabled. */
	if (FACTON_P==(FACTORINGMASK&kk)){
		kbset[nkb]->opts.factoring=1;
	} else {
		kbset[nkb]->opts.factoring=0;
	}

	/* Set term ordering. */
	if (TRMORDSTD_P==(TERMORDRMASK&kk)){
		kbset[nkb]->opts.termord=STANDARD;
	} else {
		kbset[nkb]->opts.termord=NONRECURSIVE;
	}

	/* Set literal and term ordering. */
	if (LITORDSTD_P==(LITORDRMASK&kk)){
		kbset[nkb]->opts.litord=STANDARD;
	} else if (LITORDNR_P==(LITORDRMASK&kk)){
		kbset[nkb]->opts.litord=NONRECURSIVE;
	} else {
		kbset[nkb]->opts.litord=LEXICOGRAPHIC;
	}

	/* Set function weight option. */
	if (UNIFORM_P==(FUNCWMASK&kk)){
		kbset[nkb]->opts.fweight=UNIFORM;
	} else {
		kbset[nkb]->opts.fweight=ARITY;
	}

	/* Set maximum weight option. */
	switch (MAXWMASK&kk){
		case LRSOTT_P:
			kbset[nkb]->opts.mxwghtopt=MXWEIGHTLRSOTT|(glblopts.mxwghtopt&(MXWEIGHTNUM|MXWEIGHTOFF));
			break;
		case LRSOFF_P:
			kbset[nkb]->opts.mxwghtopt=MXWEIGHTLRSOFF|(glblopts.mxwghtopt&(MXWEIGHTNUM|MXWEIGHTOFF));
			break;
		case LRSALL_P:
			kbset[nkb]->opts.mxwghtopt=MXWEIGHTLRS|(glblopts.mxwghtopt&(MXWEIGHTNUM|MXWEIGHTOFF));
			break;
		default:
			kbset[nkb]->opts.mxwghtopt=MXWEIGHTLRSDSC|(glblopts.mxwghtopt&(MXWEIGHTNUM|MXWEIGHTOFF));
			break;
	}
	kbset[nkb]->opts.maxweight=glblopts.maxweight;

	/* Initialize the learning vector for this thread. */
	kbset[nkb]->lrnvector=NULL;
	if (LEARN_P&kk){
		for (mm=0;mm<numlrnrecs;mm++){
			if ((kk&(ALGMASK|SELECTMASK|TERMORDRMASK|LITORDRMASK|FUNCWMASK))==
					((ALGMASK|SELECTMASK|TERMORDRMASK|LITORDRMASK|FUNCWMASK)&lrnrecords[mm].header)){
				kbset[nkb]->lrnvector=&lrnrecords[mm].vector[0];
				break;
			}
		}
	}

	/* Set remaining options and settings. */
	if ((glblopts.split==1)||(kbset[nkb]->opts.algorithm==glblopts.split)){
		kbset[nkb]->opts.split=1;
	} else {
		kbset[nkb]->opts.split=0;
	}
	return;
} /* SetMTParams */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Main inference thread process)  OCJ
 *
 *    This function controls the inference process for each thread.
 *    There is an instance of this function per executed thread.
 *
 *
 *  ARGUMENTS:
 *
 *    None.
 *
 *  RETURNS:
 *
 *    None.
 *
 *
 *--------------------------------------------------------------*/
void InferenceThread(void){
	auto int32_t nkb;                                    /* KB corresponding to the executing thread */
	auto int32_t slicesleft;                             /* Remaining time slices not yet run */
	auto double totaltimeleft;                           /* Total remaining time at funciton start */
	auto double slicetimeleft;                           /* Slice remaining time */
	auto struct timeval time1,time2;                     /* For time measurements */
	auto cmprefix *ptr1,*ptr2;                           /* Auxiliary pointers */
	auto int32_t ii;                                     /* Auxiliary */

	/* Initialization. */
	gettimeofday(&time1,NULL);
	time2=time1;
	nkb=omp_get_thread_num();
	slicesleft=timeslices;
	totaltimeleft=(double)(glblopts.timeout-mainkb->prstats.elapsed_time);
	slicetimeleft=totaltimeleft/slicesleft;
	kbset[nkb]->status=RUNNING;

	/* Loop for each time slice. */
	while (1){

		/* Set strategy. Return if no more strategies available. */
		omp_set_lock(&procctl.omp_lockc1);
		ii=nextstrat;
		nextstrat++;
		omp_unset_lock(&procctl.omp_lockc1);
		if (ii>=nbstrats){
			kbset[nkb]->status=UNKNOWN;
			omp_set_lock(&procctl.omp_lockc1);
			if (procctl.status==RUNNING){
				procctl.status=UNKNOWN;
			}
			omp_unset_lock(&procctl.omp_lockc1);
			break;
		}
		SetMTParams(ii,nkb);

		/* Initialize the number of formulas in the working KB with */
		/* the number of text formulas in the main KB. This way the */
		/* formula numbering in both kb's will be compatible after */
		/* loading the working KB with passive binary clauses in the */
		/* main KB. */
		kbset[nkb]->nformulas=mainkb->nformulas;

		/* Load working KB with unprocessed binary clauses in the main KB. */
		for (ptr1=mainkb->frstunproc;ptr1!=NULL;ptr1=ptr1->next){

			/* Allocate storage for the copied clause. */
			if (NULL==(ptr2=mallocmn(sizeof(cmprefix)))){
				kbset[nkb]->status=NOMEMORY;
				omp_set_lock(&procctl.omp_lockc1);
				if (procctl.status==RUNNING){
					procctl.status=NOMEMORY;
				}
				omp_unset_lock(&procctl.omp_lockc1);
				return;
			}
			memcpy(ptr2,ptr1,sizeof(cmprefix));
			if (NULL==(ptr2->part2.bin=mallocmn(binpsize+ptr1->part2.bin->size))){
				kbset[nkb]->status=NOMEMORY;
				omp_set_lock(&procctl.omp_lockc1);
				if (procctl.status==RUNNING){
					procctl.status=NOMEMORY;
				}
				omp_unset_lock(&procctl.omp_lockc1);
				MYFREE(ptr2);
				return;
			}

			/* Copy binary part of clause and clean equality data. */
			memcpy(ptr2->part2.bin,ptr1->part2.bin,binpsize+ptr1->part2.bin->size);
			ptr2->part2.bin->equflg=0;
			ptr2->part2.bin->clause=ptr2;

			/* Split clause if possible. */
			if (NOMEMORY==(ii=SplitClause(ptr2,nkb,0))){
				kbset[nkb]->status=NOMEMORY;
				omp_set_lock(&procctl.omp_lockc1);
				if (procctl.status==RUNNING){
					procctl.status=NOMEMORY;
				}
				omp_unset_lock(&procctl.omp_lockc1);
				MYFREE(ptr2->part2.bin);
				MYFREE(ptr2);
				return;
				break;
			}

			/* If clause was not split then initialize the clause symbolw set field */
			/* and add clause to KB as unprocessed. In case of an early return there */
			/* is no need to call FreeEquData() because equflg field is zero. */
			if (ii==0){
				UpdateClauseSymbolWeights(ptr2->part2.bin);
				if (NOMEMORY==AddBinClause2KB(ptr2,nkb,SETFNUMBER)){
					kbset[nkb]->status=NOMEMORY;
					omp_set_lock(&procctl.omp_lockc1);
					if (procctl.status==RUNNING){
						procctl.status=NOMEMORY;
					}
					omp_unset_lock(&procctl.omp_lockc1);
					MYFREE(ptr2->part2.bin);
					MYFREE(ptr2);
					return;
				}
			}
		}

		/* Get current time for the thread, update time left for this slice */
		/* and set thread timeout. */
		gettimeofday(&kbset[nkb]->time,NULL);
		slicetimeleft-=(kbset[nkb]->time.tv_sec-time2.tv_sec
				+((kbset[nkb]->time.tv_usec-time2.tv_usec)/1000000.0));
		kbset[nkb]->opts.timeout=slicetimeleft;

		/* Start thread algorithm. */
		switch (kbset[nkb]->opts.algorithm){
			case OTTER:
				Otter(nkb);
				break;
			case DISCOUNT:
				Discount(nkb);
				break;
		}

		/* Set time statistics for this slice. */
		gettimeofday(&kbset[nkb]->endtime,NULL);
		kbset[nkb]->prstats.elapsed_time=kbset[nkb]->endtime.tv_sec-time2.tv_sec
				+(kbset[nkb]->endtime.tv_usec-time2.tv_usec)/1000000.0;
		time2=kbset[nkb]->endtime;

		/* Update slices left and initialize time left for next slice. */
		slicetimeleft=totaltimeleft-(time2.tv_sec-time1.tv_sec+(time2.tv_usec-time1.tv_usec)/1000000.0);
		if (slicetimeleft>0){
			if (slicesleft>0){
				slicesleft--;
			}
			if (slicesleft>0){
				slicetimeleft/=slicesleft;
			}
		} else {
			omp_set_lock(&procctl.omp_lockc1);
			if ((procctl.status==RUNNING)||(procctl.status==UNKNOWN)){
				procctl.status=TIMEOUT;
			}
			omp_unset_lock(&procctl.omp_lockc1);
			break;
		}

		/* There is a solution or TIMEOUT or NOMEMORY condition. */
		if ((procctl.status!=RUNNING)&&(procctl.status!=UNKNOWN)){
			break;
		}

		/* Reinitialize KB. */
		omp_set_lock(&procctl.omp_lockc2);
		if (allcdmemory>mainkb->prstats.memory){
			mainkb->prstats.memory=allcdmemory;
		}
		AddInferenceStats(&kbset[nkb]->prstats,&mainkb->prstats);
		omp_unset_lock(&procctl.omp_lockc2);
		Initialize_hashcmp(nkb);
		Initialize_KB(nkb);
		ResetStats(&kbset[nkb]->prstats);
		kbset[nkb]->status=RUNNING;
	}

	return;
} /* InferenceThread */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Otter saturation algorithm)  OCJ
 *
 *    This function performs the Otter type saturation algorithm.
 *
 *
 *  ARGUMENTS:
 *
 *    nkb: KB for which the algorithm will be run.
 *
 *  RETURNS:
 *
 *    None.
 *
 *
 *--------------------------------------------------------------*/
void Otter(int32_t nkb){
	auto cmprefix *current;                              /* Selected current clause. */
	auto cmprefix *auxcl;                                /* Pointer to auxiliary clauses */
	auto struct timeval time1,time2;                     /* For LRS time measurements and timeout checks */
	auto double timedif;                                 /* Elapsed time in LRS cycle */
	auto double lefttime;                                /* Time available */
	auto float reachcl;                                  /* Number reachable passive clauses */
	auto uint64_t allocdmem;                             /* Allocated memory */
	auto char *ptr1;                                     /* Auxiliary pointer */
	auto int32_t ii,jj;                                  /* Auxiliary */
	#ifdef DEBUGCODE
	auto char *ptr10;                                    /* Buffer for clause in text form. */
	#endif

	/* Get initial time and initialize clause selection parameters. */
	/* The initial time for this function is in the time field of the */
	/* kbset structure. */
	time1=kbset[nkb]->time;
	kbset[nkb]->selectcnt=0;

	/* Main loop. This loop is held as far as there are passive or unprocessed */
	/* clauses or SAT queue is not empty. */
	kbset[nkb]->lrscnt=0;
	kbset[nkb]->flags&=(~RETENTIONUSED);
	while ((kbset[nkb]->frstpassive!=NULL)||(kbset[nkb]->frstunproc!=NULL)
			||(kbset[nkb]->firstsatq!=NULL)){

		/* Check memory limit. */
		#pragma omp atomic read
		allocdmem=allcdmemory;
		if (allocdmem>maxmemory){
			omp_set_lock(&procctl.omp_lockc1);
			if (procctl.status==RUNNING){
				procctl.status=NOMEMORY;
			}
			omp_unset_lock(&procctl.omp_lockc1);
			return;
		}

		/* SAT queue is not empty. */
		if (kbset[nkb]->firstsatq!=NULL){

			/* Add SAT clauses in SAT queue to SAT solver and ask SAT solver */
			/* to solve. */
			if (0==(ii=SatQ2SatSolver(nkb))){
				ii=SatSolver(nkb);
			}
			switch (ii){

				/* TIMEOUT or NOMEMORY condition. */
				case NOMEMORY:
				case TIMEOUT:
					kbset[nkb]->status=ii;
					if (ii==NOMEMORY){
						omp_set_lock(&procctl.omp_lockc1);
						if (procctl.status==RUNNING){
							procctl.status=NOMEMORY;
						}
						omp_unset_lock(&procctl.omp_lockc1);
					}
					return;
					break;

				/* Refutation found. */
				case 0:
					kbset[nkb]->status=UNSATISFIABLE;
					omp_set_lock(&procctl.omp_lockc1);
					if (procctl.status!=UNSATISFIABLE){
						procctl.status=UNSATISFIABLE;
						procctl.solvekb=nkb;
					} else {
						MYFREE(kbset[nkb]->frstprfnode);
					}
					omp_unset_lock(&procctl.omp_lockc1);
					return;
					break;

				/* A SAT model is available. */
				default:
				case 1:
					break;
			}

			/* Apply changes of current interpretation. */
			ii=ApplyModelChanges(nkb);
			if (0!=(ii=ApplyModelChanges(nkb))){
				kbset[nkb]->status=ii;
				if (ii==NOMEMORY){
					omp_set_lock(&procctl.omp_lockc1);
					if (procctl.status==RUNNING){
						procctl.status=NOMEMORY;
					}
					omp_unset_lock(&procctl.omp_lockc1);
				}
				return;
			}
		}

		/* Loop though unprocessed clauses. */
		for (auxcl=kbset[nkb]->frstunproc;auxcl!=NULL;auxcl=kbset[nkb]->frstunproc){

			/* Unlink clause. Unlink cannot return NOMEMORY because unprocessed */
			/* clauses are not indexed by discrimination trees. */
			Unlink(auxcl,nkb);

			/* Check empty clause. If the clause is empty then it has assertions so */
			/* add contradiction clause to SAT queue and iterate in the main loop. */
			/* Iterating in the main loop is achieved here by leaving the current */
			/* unprocessed clauses loop. */
			if (UNITEND==auxcl->part2.bin->formula[0]){
				if (0==(NUMBERED&auxcl->flags)){
					kbset[nkb]->nformulas++;
					auxcl->number=kbset[nkb]->nformulas;
					auxcl->flags|=NUMBERED;
				}
				if (NOMEMORY==BuildContrClause(auxcl,nkb)){
					kbset[nkb]->status=NOMEMORY;
					omp_set_lock(&procctl.omp_lockc1);
					if (procctl.status==RUNNING){
						procctl.status=NOMEMORY;
					}
					omp_unset_lock(&procctl.omp_lockc1);
					MYFREE(auxcl->part2.bin->asserts);
					MYFREE(auxcl->part2.bin);
					MYFREE(auxcl);
					return;
				}
				break;
			}

			/* Split clause if possible. If clause is split then iterate */
			/* in the current loop. In case of NOMEMORY then clause must be */
			/* freed because it is unlinked. */
			if (NOMEMORY==(ii=SplitClause(auxcl,nkb,0))){
				kbset[nkb]->status=NOMEMORY;
				omp_set_lock(&procctl.omp_lockc1);
				if (procctl.status==RUNNING){
					procctl.status=NOMEMORY;
				}
				omp_unset_lock(&procctl.omp_lockc1);
				if (auxcl->part2.bin->asserts!=NULL){
					MYFREE(auxcl->part2.bin->asserts);
				}
				MYFREE(auxcl->part2.bin);
				MYFREE(auxcl);
				return;
			}
			if (ii){
				continue;
			}

			/* Do simplifications to and from clause. */
			switch (ii=FwdSimplify(&auxcl,nkb,PASSIVE)){

				/* Check timeout and no memory. */
				case NOMEMORY:
				case TIMEOUT:
					kbset[nkb]->status=ii;
					if (ii==NOMEMORY){
						omp_set_lock(&procctl.omp_lockc1);
						if (procctl.status==RUNNING){
							procctl.status=NOMEMORY;
						}
						omp_unset_lock(&procctl.omp_lockc1);
					}
					return;
					break;

				/* Check empty clause without assertions inferred by simplification. */
				case 1:
					omp_set_lock(&procctl.omp_lockc1);
					if (procctl.status!=UNSATISFIABLE){
						procctl.status=UNSATISFIABLE;
						procctl.solvekb=nkb;
					}
					omp_unset_lock(&procctl.omp_lockc1);
					return;
					break;
			}

			/* Clause has not been locked or deleted by simplification. */
			if ((auxcl!=NULL)&&(0==(LOCKED&auxcl->flags))){

				/* Check empty clause. If the clause is empty then it has assertions so */
				/* add contradiction clause to SAT queue and iterate in the main loop. */
				/* Iterating in the main loop is achieved here by leaving the current */
				/* unprocessed clauses loop. */
				if (UNITEND==auxcl->part2.bin->formula[0]){
					if (0==(NUMBERED&auxcl->flags)){
						kbset[nkb]->nformulas++;
						auxcl->number=kbset[nkb]->nformulas;
						auxcl->flags|=NUMBERED;
					}
					if (NOMEMORY==BuildContrClause(auxcl,nkb)){
						kbset[nkb]->status=NOMEMORY;
						omp_set_lock(&procctl.omp_lockc1);
						if (procctl.status==RUNNING){
							procctl.status=NOMEMORY;
						}
						omp_unset_lock(&procctl.omp_lockc1);
						MYFREE(auxcl->part2.bin->asserts);
						MYFREE(auxcl->part2.bin);
						MYFREE(auxcl);
						return;
					}
					break;
				}

				/* Retention test checking. */
				switch (Retained(auxcl,nkb)){

					/* Not enough memory. */
					case NOMEMORY:
						kbset[nkb]->status=NOMEMORY;
						FreeEquData(auxcl->part2.bin);
						if (auxcl->part2.bin->asserts!=NULL){
							MYFREE(auxcl->part2.bin->asserts);
						}
						MYFREE(auxcl->part2.bin);
						MYFREE(auxcl);
						omp_set_lock(&procctl.omp_lockc1);
						if (procctl.status==RUNNING){
							procctl.status=NOMEMORY;
						}
						omp_unset_lock(&procctl.omp_lockc1);
						return;
						break;

					/* Clause is retained. */
					case 1:

						/* Perform backward simplification. */
						jj=BckSimplify(auxcl,nkb,PASSIVE);
						if ((jj==NOMEMORY)||(jj==TIMEOUT)){
							kbset[nkb]->status=jj;
							FreeEquData(auxcl->part2.bin);
							if (auxcl->part2.bin->asserts!=NULL){
								MYFREE(auxcl->part2.bin->asserts);
							}
							MYFREE(auxcl->part2.bin);
							MYFREE(auxcl);
							if (jj==NOMEMORY){
								omp_set_lock(&procctl.omp_lockc1);
								if (procctl.status==RUNNING){
									procctl.status=NOMEMORY;
								}
								omp_unset_lock(&procctl.omp_lockc1);
							}
							return;
						}

						/* Add clause to passive clauses.*/
						auxcl->flags&=(~(ACTIVE|PASSIVE|UNPROC|LOCKED));
						auxcl->flags|=PASSIVE;
						if (NOMEMORY==AddBinClause2KB(auxcl,nkb,SELECT|SETFNUMBER)){
							kbset[nkb]->status=NOMEMORY;
							FreeEquData(auxcl->part2.bin);
							if (auxcl->part2.bin->asserts!=NULL){
								MYFREE(auxcl->part2.bin->asserts);
							}
							MYFREE(auxcl->part2.bin);
							MYFREE(auxcl);
							omp_set_lock(&procctl.omp_lockc1);
							if (procctl.status==RUNNING){
								procctl.status=NOMEMORY;
							}
							omp_unset_lock(&procctl.omp_lockc1);
							return;
						}

						/* If BckSimplify() produced an empty clause without assertions */
						/* then assign a number to the empty clause and return. */
						if (jj){
							kbset[nkb]->status=UNSATISFIABLE;
							kbset[nkb]->nformulas++;
							kbset[nkb]->frstprfnode->ptr.Aclause->number=kbset[nkb]->nformulas;
							kbset[nkb]->frstprfnode->ptr.Aclause->flags|=NUMBERED;
							omp_set_lock(&procctl.omp_lockc1);
							if (procctl.status!=UNSATISFIABLE){
								procctl.status=UNSATISFIABLE;
								procctl.solvekb=nkb;
							} else {
								MYFREE(kbset[nkb]->frstprfnode);
							}
							omp_unset_lock(&procctl.omp_lockc1);
							return;
						}
						break;
				}
			}
		}

		/* Disposing of clauses in disposal queue. */
		for (auxcl=kbset[nkb]->lastdisp,ii=0;auxcl!=NULL;auxcl=kbset[nkb]->lastdisp,ii++){

			/* Clause is not locked. Dispose of clause. */
			if (0==(LOCKED&auxcl->flags)){
				if (NULL==(ptr1=Decompile(auxcl->part2.bin))){
					kbset[nkb]->status=NOMEMORY;
					omp_set_lock(&procctl.omp_lockc1);
					if (procctl.status==RUNNING){
						procctl.status=NOMEMORY;
					}
					omp_unset_lock(&procctl.omp_lockc1);
					return;
				}
				kbset[nkb]->lastdisp=auxcl->prevdisp;
				auxcl->prevdisp=NULL;
				auxcl->flags&=(~INDISPOSALQUEUE);
				if (NOMEMORY==Unlink(auxcl,nkb)){
					kbset[nkb]->status=NOMEMORY;
					omp_set_lock(&procctl.omp_lockc1);
					if (procctl.status==RUNNING){
						procctl.status=NOMEMORY;
					}
					omp_unset_lock(&procctl.omp_lockc1);
					return;
				}
				FreeEquData(auxcl->part2.bin);
				if (auxcl->part2.bin->asserts!=NULL){
					MYFREE(auxcl->part2.bin->asserts);
				}
				MYFREE(auxcl->part2.bin);
				auxcl->part2.text=ptr1;
				AddTxtFormula2KB(auxcl,auxcl->inference,nkb,0);

			/* If the clause is marked as LOCKED then there are locking */
			/* assertions. Unlink the clause from current queue and add */
			/* it to the LOCKED queue. If a NOMEMORY condition arises */
			/* after the clause is unlinked then it is completely deleted */
			/* because it cannot be deleted thereafter as it is unlinked. */
			} else {
				kbset[nkb]->lastdisp=auxcl->prevdisp;
				auxcl->prevdisp=NULL;
				auxcl->flags&=(~(INDISPOSALQUEUE|LOCKED));
				if (NOMEMORY==Unlink(auxcl,nkb)){
					kbset[nkb]->status=NOMEMORY;
					omp_set_lock(&procctl.omp_lockc1);
					if (procctl.status==RUNNING){
						procctl.status=NOMEMORY;
					}
					omp_unset_lock(&procctl.omp_lockc1);
					return;
				}
				auxcl->flags&=(~(ACTIVE|PASSIVE|UNPROC|LOCKED));
				auxcl->flags|=LOCKED;
				if (NOMEMORY==AddBinClause2KB(auxcl,nkb,0)){
					MYFREE(auxcl->part2.bin->lockasserts);
					if (NULL!=auxcl->part2.bin->asserts){
						MYFREE(auxcl->part2.bin->asserts);
					}
					MYFREE(auxcl->part2.bin);
					MYFREE(auxcl);
					kbset[nkb]->status=NOMEMORY;
					omp_set_lock(&procctl.omp_lockc1);
					if (procctl.status==RUNNING){
						procctl.status=NOMEMORY;
					}
					omp_unset_lock(&procctl.omp_lockc1);
					return;
				}
			}

			/* Check timeout and solution by other thread. */
	 		if (ii>=TIMEOUTCCLDSP){
				gettimeofday(&time2,NULL);
				timedif=time2.tv_sec-kbset[nkb]->time.tv_sec
						+(time2.tv_usec-kbset[nkb]->time.tv_usec)/1000000.0;
				if (timedif>=kbset[nkb]->opts.timeout){
					kbset[nkb]->status=TIMEOUT;
					return;
				}
				if ((procctl.status==UNSATISFIABLE)||(procctl.status==SATISFIABLE)){
					kbset[nkb]->status=UNKNOWN;
					return;
				}
				if ((procctl.status==NOMEMORY)||(procctl.status==TIMEOUT)){
					kbset[nkb]->status=procctl.status;
					return;
				}
				ii=-1;
	 		}
		}

		/* If SAT queue is not empty then iterate in the main loop. */
		/* It is important to place this piece of code after processing */
		/* the disposal queue to ensure that clauses marked as locked are */
		/* effectively locked. */
		if (kbset[nkb]->firstsatq!=NULL){
			continue;
		}

		/* LRS has been selected. */
		if (kbset[nkb]->opts.mxwghtopt&(MXWEIGHTLRS|MXWEIGHTLRSOTT)){

			/* Threshold has been reached. */
			kbset[nkb]->lrscnt++;
			if (kbset[nkb]->lrscnt>=LRSTHRESHOLD){

				/* Get elapsed time for last cycle and remaining available time. */
				/* Compute number of reachable passive clause. Update current time */
				/* for next cycle*/
				gettimeofday(&time2,NULL);
				timedif=time2.tv_sec-time1.tv_sec+(time2.tv_usec-time1.tv_usec)/1000000.0;
				lefttime=kbset[nkb]->opts.timeout-(time2.tv_sec-kbset[nkb]->time.tv_sec
						+(time2.tv_usec-kbset[nkb]->time.tv_usec)/1000000.0);
				reachcl=LRSCOEF*LRSTHRESHOLD*lefttime/timedif;
				gettimeofday(&time1,NULL);

				/* Not all passive clauses are reachable at the current rate. */
				if (reachcl<kbset[nkb]->prstats.nbpassive){
					if (NOMEMORY==LRSClean((int32_t)reachcl,nkb)){
						kbset[nkb]->status=NOMEMORY;
						omp_set_lock(&procctl.omp_lockc1);
						if (procctl.status==RUNNING){
							procctl.status=NOMEMORY;
						}
						omp_unset_lock(&procctl.omp_lockc1);
						return;
					}
				}

				/* Update LRS count. */
				kbset[nkb]->lrscnt=0;
			}
		}

		/* Select current clause and unlink it from KB. */
		if (NULL==(current=SelectClause(nkb))){
			break;
		}
		if (NOMEMORY==Unlink(current,nkb)){
			kbset[nkb]->status=NOMEMORY;
			omp_set_lock(&procctl.omp_lockc1);
			if (procctl.status==RUNNING){
				procctl.status=NOMEMORY;
			}
			omp_unset_lock(&procctl.omp_lockc1);
			return;
		}

		/* Debug. */
		#ifdef DEBUGCODE
		if (current->number==dbgnb){
			ptr10=NULL;
		}
		ptr10=Decompile(current->part2.bin);
		#ifdef VERBOSE
		if (num_wkbs==1){
			printf("--------------------------\nSelected clause [%d/%d]:\n%s\n",
					current->number,kbset[nkb]->nformulas,ptr10);
		}
		#endif
		MYFREE(ptr10);
		#endif

		/* Add current clause to ACTIVE queue. In case of no memory the clause must */
		/* be freed because it is unlinked and will not be freed thereafter. */
		current->flags&=(~(ACTIVE|PASSIVE|UNPROC|LOCKED));
		current->flags|=ACTIVE;
		if (NOMEMORY==AddBinClause2KB(current,nkb,SELECT)){
			FreeEquData(current->part2.bin);
			if (current->part2.bin->asserts!=NULL){
				MYFREE(current->part2.bin->asserts);
			}
			MYFREE(current->part2.bin);
			MYFREE(current);
			kbset[nkb]->status=NOMEMORY;
			omp_set_lock(&procctl.omp_lockc1);
			if (procctl.status==RUNNING){
				procctl.status=NOMEMORY;
			}
			omp_unset_lock(&procctl.omp_lockc1);
			return;
		}

		/* Perform deduction inferences. In case of an early return */
		/* current clause is freed by Infere(). */
		Infere(current,nkb);
		switch (kbset[nkb]->status){
			case NOMEMORY:
				omp_set_lock(&procctl.omp_lockc1);
				if (procctl.status==RUNNING){
					procctl.status=NOMEMORY;
				}
				omp_unset_lock(&procctl.omp_lockc1);
				return;
				break;
			case UNSATISFIABLE:
				omp_set_lock(&procctl.omp_lockc1);
				if (procctl.status!=UNSATISFIABLE){
					procctl.status=UNSATISFIABLE;
					procctl.solvekb=nkb;
				}
				omp_unset_lock(&procctl.omp_lockc1);
			/* No break */
			case UNKNOWN:
			case TIMEOUT:
				return;
				break;
		}
	}

	/* WARNING: By the moment it cannot be guaranteed that Drodi algorithms */
	/* are complete. Due to this some of the following code may be disabled. */
	/* If we are here then theory is SATISFIABLE if the algorithm used */
	/* is complete or TIMEOUT/UNKNOWN otherwise. */
	#ifdef ASSUMECOMPLETE
	if ((0==(kbset[nkb]->flags&RETENTIONUSED))&&((0==(prproflags&CLFILTERINGPRUNNED))
			||((0==(prproflags&CLFILTERINGLIMITREACHED))&&((0==((UNITEQUALITY|NONUNITEQUALITY)&pbmflags))
					||(0==(prproflags&CLFILTERINGON)))))
			&&(((MAXIMAL|SINGLENEG|MULTINEG)&kbset[nkb]->opts.select)
					&&(0==((UNITEQUALITY|NONUNITEQUALITY)&pbmflags)))){
		kbset[nkb]->status=SATISFIABLE;
		omp_set_lock(&procctl.omp_lockc1);
		if ((procctl.status!=SATISFIABLE)&&(procctl.status!=UNSATISFIABLE)){
			procctl.status=SATISFIABLE;
			procctl.solvekb=nkb;
		}
		omp_unset_lock(&procctl.omp_lockc1);
	} else {
		gettimeofday(&time2,NULL);
		timedif=time2.tv_sec-kbset[nkb]->time.tv_sec
				+(time2.tv_usec-kbset[nkb]->time.tv_usec)/1000000.0;
		if (timedif>=kbset[nkb]->opts.timeout){
			kbset[nkb]->status=TIMEOUT;
		} else {
			kbset[nkb]->status=UNKNOWN;
		}
	}
	#else
	gettimeofday(&time2,NULL);
	timedif=time2.tv_sec-kbset[nkb]->time.tv_sec
			+(time2.tv_usec-kbset[nkb]->time.tv_usec)/1000000.0;
	if (timedif>=kbset[nkb]->opts.timeout){
		kbset[nkb]->status=TIMEOUT;
	} else {
		kbset[nkb]->status=UNKNOWN;
	}
	#endif

	return;
} /* Otter */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Discount saturation algorithm)  OCJ
 *
 *    This function performs the Discount type saturation algorithm.
 *
 *
 *  ARGUMENTS:
 *
 *    nkb: KB for which the algorithm will be run.
 *
 *  RETURNS:
 *
 *    None.
 *
 *
 *--------------------------------------------------------------*/
void Discount(int32_t nkb){
	auto cmprefix *current;                              /* Selected current clause. */
	auto cmprefix *auxcl,*auxcl1;                        /* Pointers to auxiliary clause */
	auto struct timeval time1,time2;                     /* For LRS time measurements and timeout checks */
	auto double timedif;                                 /* Elapsed time in LRS cycle */
	auto double lefttime;                                /* Time available */
	auto float reachcl;                                  /* Number reachable passive clauses */
	auto uint64_t allocdmem;                             /* Allocated memory */
	auto char *ptr1;                                     /* Auxiliary pointer */
	auto int32_t ii,jj,kk;                               /* Auxiliary */
	#ifdef DEBUGCODE
	auto char *ptr10;                                    /* Buffer for clause in text form. */
	#endif

	/* Get initial time and initialize clause selection parameters. */
	/* The initial time for this function is in the time field of the */
	/* kbset structure. */
	time1=kbset[nkb]->time;
	kbset[nkb]->selectcnt=0;

	/* Main loop. */
	kbset[nkb]->lrscnt=0;
	kbset[nkb]->flags&=(~RETENTIONUSED);
	while ((kbset[nkb]->frstpassive!=NULL)||(kbset[nkb]->frstunproc!=NULL)
			||(kbset[nkb]->firstsatq!=NULL)){

		/* Check memory limit. */
		#pragma omp atomic read
		allocdmem=allcdmemory;
		if (allocdmem>maxmemory){
			omp_set_lock(&procctl.omp_lockc1);
			if (procctl.status==RUNNING){
				procctl.status=NOMEMORY;
			}
			omp_unset_lock(&procctl.omp_lockc1);
			return;
		}

		/* SAT queue is not empty. */
		if (kbset[nkb]->firstsatq!=NULL){

			/* Add SAT clauses in SAT queue to SAT solver and ask SAT solver */
			/* to solve. */
			if (0==(ii=SatQ2SatSolver(nkb))){
				ii=SatSolver(nkb);
			}
			switch (ii){

				/* TIMEOUT or NOMEMORY condition. */
				case NOMEMORY:
				case TIMEOUT:
					kbset[nkb]->status=ii;
					if (ii==NOMEMORY){
						omp_set_lock(&procctl.omp_lockc1);
						if (procctl.status==RUNNING){
							procctl.status=NOMEMORY;
						}
						omp_unset_lock(&procctl.omp_lockc1);
					}
					return;
					break;

				/* Refutation found. */
				case 0:
					kbset[nkb]->status=UNSATISFIABLE;
					omp_set_lock(&procctl.omp_lockc1);
					if (procctl.status!=UNSATISFIABLE){
						procctl.status=UNSATISFIABLE;
						procctl.solvekb=nkb;
					} else {
						MYFREE(kbset[nkb]->frstprfnode);
					}
					omp_unset_lock(&procctl.omp_lockc1);
					return;
					break;

				/* A SAT model is available. */
				default:
				case 1:
					break;
			}

			/* Apply changes of current interpretation. */
			ii=ApplyModelChanges(nkb);
			if (0!=(ii=ApplyModelChanges(nkb))){
				kbset[nkb]->status=ii;
				if (ii==NOMEMORY){
					omp_set_lock(&procctl.omp_lockc1);
					if (procctl.status==RUNNING){
						procctl.status=NOMEMORY;
					}
					omp_unset_lock(&procctl.omp_lockc1);
				}
				return;
			}
		}

		/* Loop trhough unprocessed clauses. */
		for (auxcl=kbset[nkb]->frstunproc;auxcl!=NULL;auxcl=kbset[nkb]->frstunproc){

			/* Unlink clause. Unlink cannot return NOMEMORY because */
			/* unprocessed clauses are not indexed by discrimination trees. */
			Unlink(auxcl,nkb);

			/* Check empty clause. If the clause is empty then it has assertions so */
			/* add contradiction clause to SAT queue and iterate in the main loop. */
			/* Iterating in the main loop is achieved here by leaving the current */
			/* unprocessed clauses loop. */
			if (UNITEND==auxcl->part2.bin->formula[0]){
				if (0==(NUMBERED&auxcl->flags)){
					kbset[nkb]->nformulas++;
					auxcl->number=kbset[nkb]->nformulas;
					auxcl->flags|=NUMBERED;
				}
				if (NOMEMORY==BuildContrClause(auxcl,nkb)){
					kbset[nkb]->status=NOMEMORY;
					omp_set_lock(&procctl.omp_lockc1);
					if (procctl.status==RUNNING){
						procctl.status=NOMEMORY;
					}
					omp_unset_lock(&procctl.omp_lockc1);
					MYFREE(auxcl->part2.bin->asserts);
					MYFREE(auxcl->part2.bin);
					MYFREE(auxcl);
					return;
				}
				break;
			}

			/* Split clause if possible. If clause is split then iterate */
			/* in the current loop. In case of NOMEMORY then clause must be */
			/* freed because it is unlinked. */
			if (NOMEMORY==(ii=SplitClause(auxcl,nkb,0))){
				kbset[nkb]->status=NOMEMORY;
				omp_set_lock(&procctl.omp_lockc1);
				if (procctl.status==RUNNING){
					procctl.status=NOMEMORY;
				}
				omp_unset_lock(&procctl.omp_lockc1);
				if (auxcl->part2.bin->asserts!=NULL){
					MYFREE(auxcl->part2.bin->asserts);
				}
				MYFREE(auxcl->part2.bin);
				MYFREE(auxcl);
				return;
			}
			if (ii){
				continue;
			}

			/* Do simplifications to and from clause. */
			switch (ii=FwdSimplify(&auxcl,nkb,0)){

				/* Check timeout and no memory. */
				case NOMEMORY:
				case TIMEOUT:
					kbset[nkb]->status=ii;
					if (ii==NOMEMORY){
						omp_set_lock(&procctl.omp_lockc1);
						if (procctl.status==RUNNING){
							procctl.status=NOMEMORY;
						}
						omp_unset_lock(&procctl.omp_lockc1);
					}
					return;
					break;

				/* Check empty clause without assertions inferred by simplification. */
				case 1:
					omp_set_lock(&procctl.omp_lockc1);
					if (procctl.status!=UNSATISFIABLE){
						procctl.status=UNSATISFIABLE;
						procctl.solvekb=nkb;
					}
					omp_unset_lock(&procctl.omp_lockc1);
					return;
					break;
			}

			/* Clause has not been locked or deleted by simplification. */
			if ((auxcl!=NULL)&&(0==(LOCKED&auxcl->flags))){

				/* Check empty clause. If the clause is empty then it has assertions so */
				/* add contradiction clause to SAT queue and iterate in the main loop. */
				/* Iterating in the main loop is achieved here by leaving the current */
				/* unprocessed clauses loop. */
				if (UNITEND==auxcl->part2.bin->formula[0]){
					if (0==(NUMBERED&auxcl->flags)){
						kbset[nkb]->nformulas++;
						auxcl->number=kbset[nkb]->nformulas;
						auxcl->flags|=NUMBERED;
					}
					if (NOMEMORY==BuildContrClause(auxcl,nkb)){
						kbset[nkb]->status=NOMEMORY;
						omp_set_lock(&procctl.omp_lockc1);
						if (procctl.status==RUNNING){
							procctl.status=NOMEMORY;
						}
						omp_unset_lock(&procctl.omp_lockc1);
						MYFREE(auxcl->part2.bin->asserts);
						MYFREE(auxcl->part2.bin);
						MYFREE(auxcl);
						return;
					}
					break;
				}

				/* Retention test checking. */
				switch (Retained(auxcl,nkb)){

					/* Not enough memory. */
					case NOMEMORY:
						kbset[nkb]->status=NOMEMORY;
						FreeEquData(auxcl->part2.bin);
						if (auxcl->part2.bin->asserts!=NULL){
							MYFREE(auxcl->part2.bin->asserts);
						}
						MYFREE(auxcl->part2.bin);
						MYFREE(auxcl);
						omp_set_lock(&procctl.omp_lockc1);
						if (procctl.status==RUNNING){
							procctl.status=NOMEMORY;
						}
						omp_unset_lock(&procctl.omp_lockc1);
						return;
						break;

					/* Clause is retained. Add clause to passive clauses. */
					case 1:
						auxcl->flags&=(~(ACTIVE|PASSIVE|UNPROC|LOCKED));
						auxcl->flags|=PASSIVE;
						if (NOMEMORY==AddBinClause2KB(auxcl,nkb,SELECT|SETFNUMBER)){
							kbset[nkb]->status=NOMEMORY;
							FreeEquData(auxcl->part2.bin);
							if (auxcl->part2.bin->asserts!=NULL){
								MYFREE(auxcl->part2.bin->asserts);
							}
							MYFREE(auxcl->part2.bin);
							MYFREE(auxcl);
							omp_set_lock(&procctl.omp_lockc1);
							if (procctl.status==RUNNING){
								procctl.status=NOMEMORY;
							}
							omp_unset_lock(&procctl.omp_lockc1);
							return;
						}
						break;
				}
			}
		}

		/* If SAT queue is not empty then iterate in the main loop. */
		/* There is no need to process disposal queue before the */
		/* iteration as the previous processing of unprocessed clauses */
		/* has not put any clause  in disposal queue for disposal or */
		/* locking. This is a difference with Otter() algorithm. */
		if (kbset[nkb]->firstsatq!=NULL){
			continue;
		}

		/* LRS has been selected. */
		if (kbset[nkb]->opts.mxwghtopt&(MXWEIGHTLRS|MXWEIGHTLRSDSC)){

			/* Threshold has been reached. */
			kbset[nkb]->lrscnt++;
			if (kbset[nkb]->lrscnt>=LRSTHRESHOLD){

				/* Get elapsed time for last cycle and remaining available time. */
				/* Compute number of reachable passive clause. Update current time */
				/* for next cycle*/
				gettimeofday(&time2,NULL);
				timedif=time2.tv_sec-time1.tv_sec+(time2.tv_usec-time1.tv_usec)/1000000.0;
				lefttime=kbset[nkb]->opts.timeout-(time2.tv_sec-kbset[nkb]->time.tv_sec
						+(time2.tv_usec-kbset[nkb]->time.tv_usec)/1000000.0);
				reachcl=LRSCOEF*LRSTHRESHOLD*lefttime/timedif;
				gettimeofday(&time1,NULL);

				/* Not all passive clauses are reachable at the current rate. */
				if (reachcl<kbset[nkb]->prstats.nbpassive){
					if (NOMEMORY==LRSClean((int32_t)reachcl,nkb)){
						kbset[nkb]->status=NOMEMORY;
						omp_set_lock(&procctl.omp_lockc1);
						if (procctl.status==RUNNING){
							procctl.status=NOMEMORY;
						}
						omp_unset_lock(&procctl.omp_lockc1);
						return;
					}
				}

				/* Update LRS count. */
				kbset[nkb]->lrscnt=0;
			}
		}

		/* Select current clause and unlink it from KB. */
		if (NULL==(current=SelectClause(nkb))){
			break;
		}
		if (NOMEMORY==Unlink(current,nkb)){
			kbset[nkb]->status=NOMEMORY;
			omp_set_lock(&procctl.omp_lockc1);
			if (procctl.status==RUNNING){
				procctl.status=NOMEMORY;
			}
			omp_unset_lock(&procctl.omp_lockc1);
			return;
		}

		/* Free equality ordering data. This is necessary because the clause */
		/* may be simplified in the FwdSimplify() process and then added again to */
		/* ACTIVE clauses in the Infere() process. */
		FreeEquData(current->part2.bin);

		/* Debug. */
		#ifdef DEBUGCODE
		if (current->number==dbgnb){
			ptr10=NULL;
		}
		ptr10=Decompile(current->part2.bin);
		#ifdef VERBOSE
		if (num_wkbs==1){
			printf("--------------------------\nSelected clause [%d/%d]:\n%s\n",
					current->number,kbset[nkb]->nformulas,ptr10);
		}
		#endif
		MYFREE(ptr10);
		#endif

		/* Simplify current clause by and to active clauses */
		/* but of course don't add it to PASSIVE queue. */
		/* Remember origin clause. */
		/* In case of early return the clause is freed by FwdSimplify() */
		/* or one of the functions called by FwdSimplify(). */
		auxcl1=current;
		switch (ii=FwdSimplify(&current,nkb,0)){

			/* Check timeout and no memory. */
			case NOMEMORY:
			case TIMEOUT:
				kbset[nkb]->status=ii;
				if (ii==NOMEMORY){
					omp_set_lock(&procctl.omp_lockc1);
					if (procctl.status==RUNNING){
						procctl.status=NOMEMORY;
					}
					omp_unset_lock(&procctl.omp_lockc1);
				}
				return;
				break;

				/* Check empty clause without assertions inferred by simplification. */
				case 1:
					omp_set_lock(&procctl.omp_lockc1);
					if (procctl.status!=UNSATISFIABLE){
						procctl.status=UNSATISFIABLE;
						procctl.solvekb=nkb;
					}
					omp_unset_lock(&procctl.omp_lockc1);
					return;
					break;
		}

		/* Current clause has not been locked or deleted by simplification. */
		if ((current!=NULL)&&(0==(LOCKED&current->flags))){

			/* An empty clause without assertions was inferred in current */
			/* clause simplification. */
			if (ii==1){

				/* Add current clause to UNPROC queue. */
				/* In case of no memory the clause must be freed because */
				/* it is unlinked and will not be freed thereafter. */
				current->flags&=(~(ACTIVE|PASSIVE|UNPROC|LOCKED));
				current->flags|=UNPROC;
				if (NOMEMORY==AddBinClause2KB(current,nkb,SETFNUMBER)){
					FreeEquData(current->part2.bin);
					if (current->part2.bin->asserts!=NULL){
						MYFREE(current->part2.bin->asserts);
					}
					MYFREE(current->part2.bin);
					MYFREE(current);
					kbset[nkb]->status=NOMEMORY;
					omp_set_lock(&procctl.omp_lockc1);
					if (procctl.status==RUNNING){
						procctl.status=NOMEMORY;
					}
					omp_unset_lock(&procctl.omp_lockc1);
					return;
				}

				/* Set process status and return. */
				omp_set_lock(&procctl.omp_lockc1);
				if (procctl.status!=UNSATISFIABLE){
					procctl.status=UNSATISFIABLE;
					procctl.solvekb=nkb;
				}
				omp_unset_lock(&procctl.omp_lockc1);
				return;
			}

			/* An empty clause with assertions was inferred in current clause */
			/* simplification. Add contradiction clause to SAT queue and set */
			/* current to NULL so that no additional process is performed with it. */
			if (UNITEND==current->part2.bin->formula[0]){
				if (NOMEMORY==BuildContrClause(current,nkb)){
					kbset[nkb]->status=NOMEMORY;
					omp_set_lock(&procctl.omp_lockc1);
					if (procctl.status==RUNNING){
						procctl.status=NOMEMORY;
					}
					omp_unset_lock(&procctl.omp_lockc1);
					MYFREE(current->part2.bin->asserts);
					MYFREE(current->part2.bin);
					MYFREE(current);
					return;
				}
				current=NULL;

			/* No empty clause with assertions was inferred in current clause */
			/* simplification. */
			} else {

				/* Perform backward simplification. */
				jj=BckSimplify(current,nkb,FROMPASSIVE);
				if ((jj==NOMEMORY)||(jj==TIMEOUT)){
					kbset[nkb]->status=jj;
					FreeEquData(current->part2.bin);
					if (current->part2.bin->asserts!=NULL){
						MYFREE(current->part2.bin->asserts);
					}
					MYFREE(current->part2.bin);
					MYFREE(current);
					if (ii==NOMEMORY){
						omp_set_lock(&procctl.omp_lockc1);
						if (procctl.status==RUNNING){
							procctl.status=NOMEMORY;
						}
						omp_unset_lock(&procctl.omp_lockc1);
					}
					return;
				}

				/* Add clause to ACTIVE queue. In case of no memory the clause must be */
				/* freed because it is unlinked and will not be freed thereafter. */
				current->flags&=(~(ACTIVE|PASSIVE|UNPROC|LOCKED));
				current->flags|=ACTIVE;
				if (NOMEMORY==AddBinClause2KB(current,nkb,SELECT|(auxcl1==current?0:SETFNUMBER))){
					FreeEquData(current->part2.bin);
					if (current->part2.bin->asserts!=NULL){
						MYFREE(current->part2.bin->asserts);
					}
					MYFREE(current->part2.bin);
					MYFREE(current);
					kbset[nkb]->status=NOMEMORY;
					omp_set_lock(&procctl.omp_lockc1);
					if (procctl.status==RUNNING){
						procctl.status=NOMEMORY;
					}
					omp_unset_lock(&procctl.omp_lockc1);
					return;
				}

				/* If BckSimplify() produced an empty clause without assertions */
				/* then assign a number to the empty clause and return. */
				if (jj){
					kbset[nkb]->status=UNSATISFIABLE;
					kbset[nkb]->nformulas++;
					kbset[nkb]->frstprfnode->ptr.Aclause->number=kbset[nkb]->nformulas;
					kbset[nkb]->frstprfnode->ptr.Aclause->flags|=NUMBERED;
					omp_set_lock(&procctl.omp_lockc1);
					if (procctl.status!=UNSATISFIABLE){
						procctl.status=UNSATISFIABLE;
						procctl.solvekb=nkb;
					} else {
						MYFREE(kbset[nkb]->frstprfnode);
					}
					omp_unset_lock(&procctl.omp_lockc1);
					return;
				}
			}
		}

		/* Dispose of clauses in disposal queue. */
		for (auxcl=kbset[nkb]->lastdisp,kk=0;auxcl!=NULL;auxcl=kbset[nkb]->lastdisp,kk++){

			/* Clause is not locked. Dispose of clause. */
			if (0==(LOCKED&auxcl->flags)){
				if (NULL==(ptr1=Decompile(auxcl->part2.bin))){
					kbset[nkb]->status=NOMEMORY;
					omp_set_lock(&procctl.omp_lockc1);
					if (procctl.status==RUNNING){
						procctl.status=NOMEMORY;
					}
					omp_unset_lock(&procctl.omp_lockc1);
					return;
				}
				kbset[nkb]->lastdisp=auxcl->prevdisp;
				auxcl->prevdisp=NULL;
				auxcl->flags&=(~INDISPOSALQUEUE);
				if (NOMEMORY==Unlink(auxcl,nkb)){
					kbset[nkb]->status=NOMEMORY;
					omp_set_lock(&procctl.omp_lockc1);
					if (procctl.status==RUNNING){
						procctl.status=NOMEMORY;
					}
					omp_unset_lock(&procctl.omp_lockc1);
					return;
				}
				FreeEquData(auxcl->part2.bin);
				if (auxcl->part2.bin->asserts!=NULL){
					MYFREE(auxcl->part2.bin->asserts);
				}
				MYFREE(auxcl->part2.bin);
				auxcl->part2.text=ptr1;
				AddTxtFormula2KB(auxcl,auxcl->inference,nkb,0);

			/* If the clause is marked as LOCKED then there are locking */
			/* assertions. Unlink the clause from current queue and add */
			/* it to the LOCKED queue. If a NOMEMORY condition arises */
			/* after the clause is unlinked then it is completely deleted */
			/* because it cannot be deleted thereafter as it is unlinked. */
			} else {
				kbset[nkb]->lastdisp=auxcl->prevdisp;
				auxcl->prevdisp=NULL;
				auxcl->flags&=(~(INDISPOSALQUEUE|LOCKED));
				if (NOMEMORY==Unlink(auxcl,nkb)){
					MYFREE(auxcl->part2.bin->lockasserts);
					kbset[nkb]->status=NOMEMORY;
					omp_set_lock(&procctl.omp_lockc1);
					if (procctl.status==RUNNING){
						procctl.status=NOMEMORY;
					}
					omp_unset_lock(&procctl.omp_lockc1);
					return;
				}
				auxcl->flags&=(~(ACTIVE|PASSIVE|UNPROC|LOCKED));
				auxcl->flags|=LOCKED;
				if (NOMEMORY==AddBinClause2KB(auxcl,nkb,0)){
					MYFREE(auxcl->part2.bin->lockasserts);
					if (NULL!=auxcl->part2.bin->asserts){
						MYFREE(auxcl->part2.bin->asserts);
					}
					MYFREE(auxcl->part2.bin);
					MYFREE(auxcl);
					kbset[nkb]->status=NOMEMORY;
					omp_set_lock(&procctl.omp_lockc1);
					if (procctl.status==RUNNING){
						procctl.status=NOMEMORY;
					}
					omp_unset_lock(&procctl.omp_lockc1);
					return;
				}
			}

			/* Check timeout and solution by other thread. */
			if (kk>=TIMEOUTCCLDSP){
				gettimeofday(&time2,NULL);
				timedif=time2.tv_sec-kbset[nkb]->time.tv_sec
						+(time2.tv_usec-kbset[nkb]->time.tv_usec)/1000000.0;
				if (timedif>=kbset[nkb]->opts.timeout){
					kbset[nkb]->status=TIMEOUT;
					return;
				}
				if ((procctl.status==UNSATISFIABLE)||(procctl.status==SATISFIABLE)){
					kbset[nkb]->status=UNKNOWN;
					return;
				}
				if ((procctl.status==NOMEMORY)||(procctl.status==TIMEOUT)){
					kbset[nkb]->status=procctl.status;
					return;
				}
				kk=-1;
			}
		}

		/* Current clause has not been locked or deleted by simplification. */
		/* Perform deduction inferences. In case of an early return */
		/* the current clause is freed in Infere(). */
		if ((current!=NULL)&&(0==(LOCKED&current->flags))){
			Infere(current,nkb);
			switch (kbset[nkb]->status){
				case NOMEMORY:
					omp_set_lock(&procctl.omp_lockc1);
					if (procctl.status==RUNNING){
						procctl.status=NOMEMORY;
					}
					omp_unset_lock(&procctl.omp_lockc1);
					return;
					break;
				case UNSATISFIABLE:
					omp_set_lock(&procctl.omp_lockc1);
					if (procctl.status!=UNSATISFIABLE){
						procctl.status=UNSATISFIABLE;
						procctl.solvekb=nkb;
					}
					omp_unset_lock(&procctl.omp_lockc1);
				/* No break */
				case UNKNOWN:
				case TIMEOUT:
					return;
					break;
			}
		}
	}

	/* WARNING: By the moment it cannot be guaranteed that Drodi algorithms */
	/* are complete. Due to this some of the following code may be disabled. */
	/* If we are here then theory is SATISFIABLE if the algorithm used */
	/* is complete or UNKNOWN otherwise. */
	#ifdef ASSUMECOMPLETE
	if ((0==(kbset[nkb]->flags&RETENTIONUSED))&&((0==(prproflags&CLFILTERINGPRUNNED))
			||((0==(prproflags&CLFILTERINGLIMITREACHED))
					&&((0==((UNITEQUALITY|NONUNITEQUALITY)&pbmflags))||(0==(prproflags&CLFILTERINGON)))))
			&&(((MAXIMAL|SINGLENEG|MULTINEG)&kbset[nkb]->opts.select)
					&&(0==((UNITEQUALITY|NONUNITEQUALITY)&pbmflags)))){
		kbset[nkb]->status=SATISFIABLE;
		omp_set_lock(&procctl.omp_lockc1);
		if ((procctl.status!=SATISFIABLE)&&(procctl.status!=UNSATISFIABLE)){
			procctl.status=SATISFIABLE;
			procctl.solvekb=nkb;
		}
		omp_unset_lock(&procctl.omp_lockc1);
	} else {
		kbset[nkb]->status=UNKNOWN;
	}
	#else
	kbset[nkb]->status=UNKNOWN;
	#endif

	return;
} /* Discount */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Select current clause from passive clauses)  OCJ
 *
 *    This function selects a "current" clause from the set of
 *    passive clauses in a given KB.
 *
 *    The selection algorithm is based on the "age" and "weight" queues
 *    of the KB and on an age to weight ratio that ensures fairness.
 *
 *
 *  ARGUMENTS:
 *
 *    nkb: KB for which the algorithm will be run.
 *
 *  RETURNS:
 *
 *    Pointer to selected "current" clause or NULL if no more passive
 *    clauses available.
 *
 *
 *--------------------------------------------------------------*/
cmprefix *SelectClause(int32_t nkb){
	auto int32_t jj;                                     /* Auxiliary */

	/* There are passive clauses. */
	if (kbset[nkb]->frstpassive!=NULL){

		/* Get queue index with learning disabled. */
		if (kbset[nkb]->lrnvector==NULL){
			jj=slctordr[kbset[nkb]->selectcnt];
			kbset[nkb]->selectcnt++;
			if (kbset[nkb]->selectcnt==slctordrsize){
				kbset[nkb]->selectcnt=0;
			}

		/* Get queue index with learning enabled. */
		} else {
			jj=slctordr2[kbset[nkb]->selectcnt];
			kbset[nkb]->selectcnt++;
			if (kbset[nkb]->selectcnt==slctordrsize2){
				kbset[nkb]->selectcnt=0;
			}
		}

		/* Select by weight. */
		if (jj<WEIGHTQUEUES){
			#ifdef PREFERPROCESSED
			kbset[nkb]->frstglweight[jj]->flags|=SELECTEDCLAUSE;
			#endif
			kbset[nkb]->frstglweight[jj]->flags|=PICKED;
			return(kbset[nkb]->frstglweight[jj]);

		/* Select by age. Update counter, indicate that clause has been selected and return. */
		} else {
			#ifdef PREFERPROCESSED
			kbset[nkb]->frstpassive->flags|=SELECTEDCLAUSE;
			#endif
			kbset[nkb]->frstpassive->flags|=PICKED;
			return(kbset[nkb]->frstpassive);
		}
	}

	return(NULL);
} /* SelectClause */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Check if clause is retained)  OCJ
 *
 *    This function checks if a clause is retained. The clause is assumed
 *    to be unlinked from KB.
 *
 *    If the clause is not retained then it is converted to text and the
 *    binary (non text) data is freed.
 *
 *
 *  ARGUMENTS:
 *
 *    clause: Pointer to clause.
 *    nkb: KB for which the algorithm will be run.
 *
 *  RETURNS:
 *
 *    0 -> Clause was not retained.
 *    1 -> Clause is retained.
 *
 *
 *--------------------------------------------------------------*/
int32_t Retained(cmprefix *clause,int32_t nkb){
	auto char *ptr1;                                     /* Auxiliary pointer */

	/* Check clause weight. The weight is only checked for the last */
	/* weight queue. */
	ClauseWeight(clause,nkb);
	if (clause->part2.bin->clweight[WEIGHTQUEUES-1]>=kbset[nkb]->opts.maxweight){

		/* Convert clause to text form. Due to Avatar architecture the same */
		/* clause can go as unprocessed several times, so an unprocessed */
		/* clause is not necessarily new and therefore can be part of the */
		/* theorem proof. */
		if (NULL==(ptr1=Decompile(clause->part2.bin))){
			return(NOMEMORY);
		}
		kbset[nkb]->flags|=RETENTIONUSED;
		FreeEquData(clause->part2.bin);
		if (clause->part2.bin->asserts!=NULL){
			MYFREE(clause->part2.bin->asserts);
		}
		MYFREE(clause->part2.bin);
		clause->part2.text=ptr1;
		AddTxtFormula2KB(clause,clause->inference,nkb,SETFNUMBER);
		return(0);
	}

	return(1);
} /* Retained */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Compute clause weight)  OCJ
 *
 *    This function computes the weight of a clause in compiled binary
 *    format. The following definitions are used:
 *
 *    Distance to conjecture:
 *    ----------------------
 *    - Clauses obtained from conjecture or negated conjecture in the
 *      pre-processing have distance=0.
 *    - Clauses obtained in the prep-rocessing not from conjecture or
 *      negated conjecture have distance=MAXDIST2CONJ.
 *    - If a clause is simplified the resulting clause has the same
 *      distance as the original clause.
 *    - If a clause is obtained by other inferences from parent clauses
 *      with distances m and n then the clause distance is
 *      MIN(MAXDIST2CONJ,1+MIN(m,n)).
 *
 *    Symbol weights:
 *    --------------
 *    There are two types of symbol weight, stored in the symbolw[] set
 *    field of the binprefix structure:
 *    - symbolw[0] is computed assigning weights VARWEIGHT, PREDWEIGHT
 *      or FUNCWEIGHT to each variable, predicate or function symbol.
 *    - symbol[1] is computed as symbol[0] but if a symbol doesn't
 *      appear in the conjecture then the symbol weight is multiplied
 *      by the appropriate factor, depending on the kind of symbol
 *      and the value of kbset[nkb]->opts.factoridx, so getting heavier.
 *      See PNOTINCONJFACTOR1, ENOTINCONJFACTOR1 and FNOTINCONJFACTOR1
 *      definitions in global.h.
 *
 *    In addition, for both cases above if a symbol is in a positive
 *    literal then its weights are multiplied by POSLITERALFACTOR1 or
 *    POSLITERALFACTOR2, depending on the value of kbset[nkb]->opts.factoridx.
 *
 *    The following weights are calculated:
 *
 *    Weight 0:
 *    --------
 *    - This weight is only used if LEARNFROM option is enabled and there is
 *      a learn vector available for the thread. The weight is computed as in
 *      weight 1, multiplied by a factor usually less than one and added
 *      BADCLAUSEWEIGHT if the clause is not classified as good by the learning
 *      process.
 *
 *    Weight 1:
 *    --------
 *    - Clauses with distance=0 have weight=0.
 *    - Other clauses obtained in the pre-processing have weight=1. These
 *      clauses have inference=CLAUSIFY and distance=MAXDIST2CONJ.
 *    - For the rest of the clauses weight=symbolw[0]+10. This weight is bigger
 *      than or equal to 20.
 *
 *    Weight 2:
 *    --------
 *    - If the distance of clause to conjecture is less than MAXDIST2CONJ
 *      then then weight=symbolw[0]-10. Otherwise weight=symbolw[0].
 *
 *    Weight 3:
 *    --------
 *    - The weight is computed as weight=symbolw[1]-10.
 *
 *    Weight 4:
 *    --------
 *    - The weight is computed as weight=symbolw[0]-10.
 *
 *   VERY IMPORTANT: lowering of maximum clause weight allowed for clause
 *   retention is done in LRSClean() function based only on the last weight
 *   set (weight 4).
 *
 *
 *  ARGUMENTS:
 *
 *    clause: Pointer to clause.
 *    nkb: KB for which the algorithm will be run.
 *
 *  RETURNS:
 *
 *    None.
 *
 *
 *--------------------------------------------------------------*/
void ClauseWeight(cmprefix *clause,int32_t nkb){
	auto double *clvector;                               /* Pointer to clause learning vector */
	auto double dd;                                      /* Auxiliary */
	auto int32_t ii;                                     /* Auxiliary */

	/* Return if clause weight has been already set. */
	if (clause->flags&WEIGHTED){
		return;
	}

	/* Clause weight 1. */
	if (clause->part2.bin->dist2conj==0){
		clause->part2.bin->clweight[1]=0;
	} else if (clause->inference==CLAUSIFY){
		clause->part2.bin->clweight[1]=1;
	} else {
		clause->part2.bin->clweight[1]=(int)(clause->part2.bin->symbolw[0]);
	}

	/* Clause weight 2. */
	if (clause->part2.bin->dist2conj<MAXDIST2CONJ){
		if (clause->part2.bin->symbolw[0]>=10.0){
			clause->part2.bin->clweight[2]=(int)(clause->part2.bin->symbolw[0])-10;
		} else {
			clause->part2.bin->clweight[2]=0;
		}
	} else {
		clause->part2.bin->clweight[2]=(int)(clause->part2.bin->symbolw[0]);
	}

	/* Clause weight 3. */
	if (clause->part2.bin->symbolw[1]>=10.0){
		clause->part2.bin->clweight[3]=(int)(clause->part2.bin->symbolw[1]-10.0);
	} else {
		clause->part2.bin->clweight[3]=0;
	}

	/* Clause weight 4. */
	if (clause->part2.bin->symbolw[0]>=10.0){
		clause->part2.bin->clweight[4]=(int)(clause->part2.bin->symbolw[0]-10.0);
	} else {
		clause->part2.bin->clweight[4]=0;
	}

	/* Clause weight 0. This is done last because because it uses one of the */
	/* previous weights. Currently it uses weight 1. If this is modified then */
	/* please modify also the global comments. */
	if (NULL!=kbset[nkb]->lrnvector){
		clause->part2.bin->clweight[0]=WEIGHT0FACTOR*clause->part2.bin->clweight[1];
		clvector=GetClauseVector(clause);
		if (clvector!=NULL){
			for (ii=0,dd=0.0;ii<VECTORLENGTH;ii++){
				dd+=(clvector[ii]*kbset[nkb]->lrnvector[ii]);
			}
			MYFREE(clvector);
			if (dd>0.0){
				clause->flags|=GOODCHOICE;
			}
			if (0==(GOODCHOICE&clause->flags)){
				clause->part2.bin->clweight[0]+=BADCLAUSEWEIGHT;
			}
		}
	}

	/* Indicate that clause has been weighted and return. */
	clause->flags|=WEIGHTED;
	return;
} /* ClauseWeight */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Delete unreachable passive clauses)  OCJ
 *
 *    This function deletes unreachable passive clauses for a KB
 *    given the number of reachable clauses.
 *
 *    VERY IMPORTANT: lowering of maximum clause weight allowed for clause
 *    retention is done based only on the last weight set (weight 4).
 *
 *
 *  ARGUMENTS:
 *
 *    reachcl: Number of reachable passive clauses.
 *    nkb: KB for which the algorithm will be run.
 *
 *  RETURNS:
 *
 *    0 if OK or NOMEMORY.
 *
 *
 *--------------------------------------------------------------*/
int32_t LRSClean(int32_t reachcl,int32_t nkb){
	auto cmprefix *wnxtclause[WEIGHTQUEUES];             /* Pointer to next clause in each weight queue to be signed by weight */
	auto cmprefix *anxtclause;                           /* Pointer to next clause to be signed by age */
	auto cmprefix *auxcl1,*auxcl2;                       /* Auxiliary clauses */
	auto int32_t selectcnt;                              /* Current number of selected clauses from queue indexed by selectidx */
	auto char *ptr1;                                     /* Auxiliary pointer */
	auto int32_t ii,jj,kk;                               /* Auxiliary */

	/* Indicate that LRS has been used. */
	kbset[nkb]->flags|=RETENTIONUSED;

	/* Get first clause in weight and age queues. */
	selectcnt=kbset[nkb]->selectcnt;
	for (ii=0;ii<WEIGHTQUEUES;ii++){
		wnxtclause[ii]=kbset[nkb]->frstglweight[ii];
	}
	anxtclause=kbset[nkb]->frstpassive;

	/* Sign the reachable passive clauses. */
	kbset[nkb]->signature++;
	for (ii=0;ii<reachcl;ii++){

		/* Select by weight. */
		if (NULL==kbset[nkb]->lrnvector){
			kk=slctordr[selectcnt];
		} else {
			kk=slctordr2[selectcnt];
		}
		if (kk<WEIGHTQUEUES){

			/* Jump over already signed clauses. */
			for (;wnxtclause[kk]->part2.bin->signature==kbset[nkb]->signature;
					wnxtclause[kk]=wnxtclause[kk]->part2.bin->nextweight[kk]){
			}

			/* Sign clause and update index and counter. */
			wnxtclause[kk]->part2.bin->signature=kbset[nkb]->signature;
			wnxtclause[kk]=wnxtclause[kk]->part2.bin->nextweight[kk];
			selectcnt++;
			if (NULL==kbset[nkb]->lrnvector){
				if (selectcnt==slctordrsize){
					selectcnt=0;
				}
			} else {
				if (selectcnt==slctordrsize2){
					selectcnt=0;
				}
			}

		/* Select by age. */
		} else {

			/* Jump over already signed clauses. */
			for (;anxtclause->part2.bin->signature==kbset[nkb]->signature;
					anxtclause=anxtclause->next){
			}

			/* Sign clause and update index and counter. */
			anxtclause->part2.bin->signature=kbset[nkb]->signature;
			anxtclause=anxtclause->next;
			selectcnt++;
			if (NULL==kbset[nkb]->lrnvector){
				if (selectcnt==slctordrsize){
					selectcnt=0;
				}
			} else {
				if (selectcnt==slctordrsize2){
					selectcnt=0;
				}
			}
		}
	}

	/* Dispose of unsigned clauses. */
	for (auxcl1=wnxtclause[0];auxcl1!=NULL;auxcl1=auxcl2){
		auxcl2=auxcl1->part2.bin->nextweight[0];
		if (auxcl1->part2.bin->signature!=kbset[nkb]->signature){
			if (NULL==(ptr1=Decompile(auxcl1->part2.bin))){
				return(NOMEMORY);
			}
			if (NOMEMORY==Unlink(auxcl1,nkb)){
				return(NOMEMORY);
			}
			FreeEquData(auxcl1->part2.bin);
			if (auxcl1->part2.bin->asserts!=NULL){
				MYFREE(auxcl1->part2.bin->asserts);
			}
			MYFREE(auxcl1->part2.bin);
			auxcl1->part2.text=ptr1;
			AddTxtFormula2KB(auxcl1,auxcl1->inference,nkb,0);
		}
	}

	/* Update current maximum clause weight to the maximum existing weight */
	/* in the last weight queue. */
	for (auxcl1=kbset[nkb]->lstweight[WEIGHTQUEUES-1][MAXWEIGHT],jj=MAXWEIGHT;auxcl1==NULL;
			auxcl1=kbset[nkb]->lstweight[WEIGHTQUEUES-1][--jj]){
	}
	if (kbset[nkb]->opts.maxweight>auxcl1->part2.bin->clweight[WEIGHTQUEUES-1]){
		kbset[nkb]->opts.maxweight=auxcl1->part2.bin->clweight[WEIGHTQUEUES-1];
	}

	return(0);
} /* LRSClean */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Set select order for clause selection queues)  OCJ
 *
 *    This function initializes the sets pointed by slctordr and
 *    slctord2 pointers. The sets are allocated or reallocated if
 *    necessary. They contain WEIGHTQUEUES+1 elements, one for each
 *    extraction from the age and weight queues during a round-robin
 *    clause selection cycle. Each element indicates from what queue
 *    the extraction must be done.
 *
 *    The slctordrsize and slctordsize2 global variables are also
 *    initialized.
 *
 *
 *  ARGUMENTS:
 *
 *    None.
 *
 *  RETURNS:
 *
 *    0 if OK or NOMEMORY.
 *
 *
 *--------------------------------------------------------------*/
int32_t SetClSelectOrder(void){
	auto int32_t *ptr1;                                  /* Auxiliary pointer */
	auto int32_t ii,jj,kk,mm;                            /* Auxiliary */
	#ifdef PROGCLAUSESELECT
	auto float ff,gg,hh;                                 /* Auxiliary */
	#endif

	/* Count the number of extractions. */
	for (ii=jj=mm=0;ii<=WEIGHTQUEUES;ii++){
		jj+=queuerts[ii];
		mm+=queuerts2[ii];
	}

	/* Allocate or reallocate slctordr and slctord2 set if necessary */
	/* and set slctordrsize and slctordrsize2. */
	if (slctordr==NULL){
		if (NULL==(slctordr=mallocmn(jj*sizeof(*slctordr)))){
			return(NOMEMORY);
		}
		slctordrsize=jj;
	} else if (slctordrsize!=jj){
		if (NULL==(ptr1=reallocmn(slctordr,jj*sizeof(*slctordr)))){
			return(NOMEMORY);
		}
		slctordr=ptr1;
		slctordrsize=jj;
	}
	if (slctordr2==NULL){
		if (NULL==(slctordr2=mallocmn(mm*sizeof(*slctordr2)))){
			MYFREE(slctordr);
			slctordr=NULL;
			return(NOMEMORY);
		}
		slctordrsize2=mm;
	} else if (slctordrsize2!=mm){
		if (NULL==(ptr1=reallocmn(slctordr2,mm*sizeof(*slctordr2)))){
			MYFREE(slctordr);
			slctordr=NULL;
			return(NOMEMORY);
		}
		slctordr2=ptr1;
		slctordrsize2=mm;
	}

	#ifdef PROGCLAUSESELECT
	/* Set slctordr elements by extracting progressively from each queue. */
	/* First allocate a set of WEIGHTQUEUES+1 integers to store the */
	/* number of extractions done so far from each queue. */
	if (NULL==(ptr1=allocmn(WEIGHTQUEUES+1,sizeof(int32_t)))){
		return(NOMEMORY);
	}
	memset(ptr1,0,(WEIGHTQUEUES+1)*sizeof(int32_t));

	/* Loop through all queue extractions */
	/* when learning process disabled. */
	for (kk=0;kk<slctordrsize;kk++){

		/* Loop through all candidate queues to get extraction queue. */
		ff=(float)(kk+1.0)/(float)slctordrsize;
		gg=(ff*queuerts[wqdeplord[0]])-ptr1[wqdeplord[0]];
		for (ii=1,jj=wqdeplord[0];ii<WEIGHTQUEUES;ii++){
			hh=(ff*queuerts[wqdeplord[ii]])-ptr1[wqdeplord[ii]];
			if (gg<hh){
				jj=wqdeplord[ii];
				gg=hh;
			}
		}
		hh=(ff*queuerts[WEIGHTQUEUES])-ptr1[WEIGHTQUEUES];
		if (gg<hh){
			jj=WEIGHTQUEUES;
			gg=hh;
		}
		slctordr[kk]=jj;
		ptr1[jj]++;
	}

	/* Loop through all queue extractions. */
	/* when learning process enabled. */
	memset(ptr1,0,(WEIGHTQUEUES+1)*sizeof(int32_t));
	for (kk=0;kk<slctordrsize2;kk++){

		/* Loop through all candidate queues to get extraction queue. */
		ff=(float)(kk+1.0)/(float)slctordrsize2;
		gg=(ff*queuerts2[wqdeplord2[0]])-ptr1[wqdeplord2[0]];
		for (ii=1,jj=wqdeplord2[0];ii<WEIGHTQUEUES;ii++){
			hh=(ff*queuerts2[wqdeplord2[ii]])-ptr1[wqdeplord2[ii]];
			if (gg<hh){
				jj=wqdeplord2[ii];
				gg=hh;
			}
		}
		hh=(ff*queuerts2[WEIGHTQUEUES])-ptr1[WEIGHTQUEUES];
		if (gg<hh){
			jj=WEIGHTQUEUES;
			gg=hh;
		}
		slctordr2[kk]=jj;
		ptr1[jj]++;
	}
	MYFREE(ptr1);

	#else
	/* Set slctordr elements by depleting each queue one after the other */
	/* when learning process disabled. */
	for (ii=jj=0;ii<WEIGHTQUEUES;ii++){
		for (kk=0;kk<queuerts[wqdeplord[ii]];jj++,kk++){
			slctordr[jj]=wqdeplord[ii];
		}
	}
	for (;jj<slctordrsize;jj++){
		slctordr[jj]=WEIGHTQUEUES;
	}

	/* Set slctordr elements by depleting each queue one after the other */
	/* when learning process enabled. */
	for (ii=jj=0;ii<WEIGHTQUEUES;ii++){
		for (kk=0;kk<queuerts2[wqdeplord2[ii]];jj++,kk++){
			slctordr2[jj]=wqdeplord2[ii];
		}
	}
	for (;jj<slctordrsize2;jj++){
		slctordr2[jj]=WEIGHTQUEUES;
	}
	#endif

	return(0);
} /* SetClSelectOrder */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Perform problem pre-processing)  OCJ
 *
 *    This function performs the pre-processing needed before the
 *    start of the saturation process or SAT stand alone execution:
 *    - Convert formulas to CNF standard form.
 *    - Perform global pre-processing.
 *    - Update the HORN flag of unprocessed clauses and pbmflags
 *      global variable.
 *
 *
 *  ARGUMENTS:
 *
 *    flag: 1 if called from Saturate, 2 if called from Clausify,
 *          0 otherwise.
 *
 *  RETURNS:
 *
 *    0 if OK, otherwise NOMEMORY or TIMEOUT.
 *
 *
 *--------------------------------------------------------------*/
int32_t Preprocess(int32_t flag){
	auto char *buffer;                               /* Buffer for CNF formula */
	auto int32_t size;                               /* Buffer size in bytes */
	auto cmprefix *ptr1,*ptr2;                       /* Auxiliary pointers */
	auto uint8_t *ptr3;                              /* Auxiliar pointer */
	auto uint64_t ii,jj,kk,nn,mm;                    /* Auxiliary */

	/* Allocate space for CNF conversion. */
	if (NULL==(buffer=mallocmn(2*STR_CHUNK_SIZE))){
		return(NOMEMORY);
	}
	size=2*STR_CHUNK_SIZE;

	/* Loop through text formulas in the main KB. */
	if ((verbose)&&(pgmmode==INTERACTIVE)){
		printf("Converting formulas to CNF...\n");
	}
	for (ptr1=mainkb->firsttxt;ptr1!=NULL;ptr1=ptr1->next){

		/* Formula is of type INPUT, NEGCONJ or PREDICATEDEF. */
		if ((ptr1->inference==INPUT)||(ptr1->inference==NEGCONJ)||(ptr1->inference==PREDICATEDEF)){

			/* Text formula pre-processing. Pre-processing is repeated in */
			/* case of premature return due to miniscoping time limit */
			/* exceeded. */
			if (NULL==(ptr2=PreprocessFormula(ptr1,&buffer,&size,1))){
				if ((procctl.status!=NOMEMORY)&&(procctl.status!=TIMEOUT)){
					if (NULL==(ptr2=PreprocessFormula(ptr1,&buffer,&size,0))){
						break;
					}
				} else {
					break;
				}
			}

			/* Formula is $false, refutation found. */
			if (0==strcmp("$false",ptr2->part2.text)){
				procctl.status=UNSATISFIABLE;
				procctl.solvekb=num_wkbs;
				if (NULL!=(mainkb->frstprfnode=mallocmn(sizeof(proofnode)))){
					mainkb->frstprfnode->type=ACLAUSE;
					mainkb->frstprfnode->ptr.Aclause=ptr2;
				}

			/* Formula is not $false or $true, set FROMCONJECTURE flag */
			/* if appropriate and compile CNF formula. */
			} else if (0!=strcmp("$true",ptr2->part2.text)){
				if (ptr1->inference==NEGCONJ){
					ptr2->flags|=FROMCONJECTURE;
				} else {
					ptr2->flags|=(FROMCONJECTURE&ptr1->flags);
				}
				if (NOMEMORY==CompileCNF(buffer,ptr2)){
					procctl.status=NOMEMORY;
					break;;
				}
			}
		}
	}

	/* Free CNF conversion memory. Free also main KB tree memory */
	/* as main KB trees are no longer needed. */
	MYFREE(buffer);
	FreeTreeMemory(num_wkbs);

	/* Timeout or not enough memory from PreprocessFormula(). */
	if ((procctl.status==TIMEOUT)||(procctl.status==NOMEMORY)){
		return(procctl.status);
	}

	/* Return if called from Clausify(). */
	if (flag==2){
		return(0);
	}

	/* Update the HORN flag in each clause and UNITEQUALITY and UNITCLAUSE */
	/* flags of pbmflags global variable. */
	/* Loop through unprocessed clauses. */
	/* ii=1 if there are horn clauses, otherwise ii=0. */
	/* jj=1 if there are non horn clauses, otherwise jj=0. */
	/* nn=1 if there are equalities, otherwise nn=0. */
	ii=jj=nn=0;
	for (ptr1=mainkb->frstunproc;ptr1!=NULL;ptr1=ptr1->next){

		/* Update UNITCLAUSE and UNITEQUALITY flags. */
		if (ptr1->part2.bin->literals>1){
			pbmflags&=(~UNITCLAUSE);
		} else if (EQUALITY==((NEGATED|EQUALITY)&ptr1->part2.bin->formula[0])){
			pbmflags|=UNITEQUALITY;
		}

		/* Loop through clause items. Set kk to the number of positive */
		/* literals and remove the GROUND flag if clause is not ground. */
		for (ptr3=&ptr1->part2.bin->formula[0],kk=0;(*ptr3)!=UNITEND;
				ptr3=NextItem(ptr3,IMMED)){
			if ((PREDICATE|EQUALITY)&ptr3[0]){
				if (0==(NEGATED&ptr3[0])){
					kk++;
				}
				if (EQUALITY&ptr3[0]){
					nn=1;
				}
			} else if (VARIABLE&ptr3[0]){
				pbmflags&=(~GROUND);
			}
		}

		/* Update ii, jj and clause HORN flag if appropriate. */
		if (kk>1){
			jj=1;
		} else {
			ptr1->flags|=HORN;
			ii=1;
		}
	}

	/* Perform global pre-processing. */
	mm=GlobalPreprocess();

	/* Update NONUNITEQUALITY, HORN and WITHHORNCLAUSES in pbmflags global variable. */
	if (nn&&(0==(pbmflags&UNITEQUALITY))){
		pbmflags|=NONUNITEQUALITY;
	}
	if (ii){
		if (jj){
			pbmflags|=WITHHORNCLAUSES;
		} else {
			pbmflags|=HORN;
		}
	}

	/* Accumulate preprocessing elapsed time. */
	gettimeofday(&mainkb->endtime,NULL);
	mainkb->prstats.elapsed_time+=(mainkb->endtime.tv_sec-mainkb->time.tv_sec
			+(mainkb->endtime.tv_usec-mainkb->time.tv_usec)/1000000.0);

	/* Print preprocessing statistics. */
	if ((verbose)&&(pgmmode==INTERACTIVE)){
		printf("\nPreprocessing statistics:\n");
		PrintStatData(&mainkb->prstats,0);
		if (mm){
			printf("%ld clauses deleted by relevance filtering\n",mm);
		}
	}
	if (pgmmode==INTERACTIVE){
		if (flag==1){
			printf("Problems processed: %d, problems solved: %d, now working on problem number %d\n",
					tot_pbms,solved_pbms,tot_pbms+1);
		} else {
			printf("Problems processed: %d, now working on problem number %d\n",tot_pbms,tot_pbms+1);
		}
	}

	return(0);
} /* Preprocess */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Perform problem global pre-processing)  OCJ
 *
 *    This function select clauses to be used in the saturation process.
 *    Pruning starts with the function and predicate symbols in the goal.
 *    Every axiom that shares such a symbol is considered relevant. Symbols
 *    in relevant axioms become relevant themselves. The process is then
 *    repeated for the time specified in the program options. This relevance
 *    pruning is complete in the non-equational case if allowed to reach a
 *    fixed point. It only provides a relatively coarse measure, however.
 *
 *    If the global variable prproflags has the CLFILTERINGITERLIMIT flag
 *    then the global variable ppthrshld contains the maximun number of
 *    iterations allowed. Otherwise the global variable ppthrshld contains
 *    the maximum allowed time for the algorithm.
 *
 *    A Sine algorithm can be added in the future for a more fine clause
 *    selection.
 *
 *
 *  ARGUMENTS:
 *
 *    None.
 *
 *  RETURNS:
 *
 *    Number of deleted clauses by the relevance filter.
 *
 *
 *--------------------------------------------------------------*/
uint64_t GlobalPreprocess(void){
	struct timeval sttime,ctime;                         /* Function start and current times */
	auto double timedif;                                 /* Function elapsed time */
	auto cmprefix *ptr1,*ptr2;                           /* Auxiliary pointers */
	auto uint8_t *ptr3;                                  /* Auxiliary pointer */
	auto int32_t ii,jj,mm,nn;                            /* Auxiliary */
	auto uint64_t kk;                                    /* Auxiliary */

	/* Clear CLFILTERINGLIMITREACHED and CLFILTERINGPRUNNED flag from previous jobs. */
	prproflags&=~(CLFILTERINGLIMITREACHED|CLFILTERINGPRUNNED);

	/* Return if no clause filtering must be done. */
	if ((0==(pbmflags&(EXISTCONJ|EXISTNEGCONJ)))||(prproflags&CLFILTERINGOFF)
			||((prproflags&CLFILTERINGNONEQU)&&((UNITEQUALITY|NONUNITEQUALITY)&pbmflags))){
		return(0);
	}

	/* Get start time if necessary. */
	if ((ppthrshld>0.0)&&(0==(prproflags&CLFILTERINGITERLIMIT))){
		gettimeofday(&sttime,NULL);
	}

	/* Set ptr2 to first main KB unprocessed clause and loop until */
	/* no more clauses are flagged as kept or a timeout occurs. */
	ptr2=mainkb->frstunproc;
	mm=0; /* No clause coming from negated has been processed yet. */
	for (ii=nn=1;ii;nn++){
		ii=0;

		/* Loop through main KB unprocessed clauses. The pointer ptr2 */
		/* is set to first non kept clause in each iteration. */
		for (ptr1=ptr2;ptr1!=NULL;ptr1=ptr1->next){
			if (ptr1==ptr2){
				ptr2=NULL;
			}

			/* Clause is not flagged as kept. */
			if (0==(KEPTCLAUSE&ptr1->flags)){

				/* Clause is coming from a negated conjecture. */
				if (ptr1->part2.bin->dist2conj==0){
					jj=mm=1;

				/* Clause is not coming from a negated conjecture. */
				} else {

					/* At least one clause coming from conjecture has been processed. */
					if (mm){

						/* Loop through clause items and check if any of its */
						/* predicate or function symbols are flagged as RELEVANT. */
						for (ptr3=&ptr1->part2.bin->formula[0],jj=0;(jj==0)&&(ptr3[0]!=UNITEND);ptr3=NextItem(ptr3,IMMED)){
							if (((PREDICATE|FUNCTION)&ptr3[0])&&(RELEVANT&((symbol *)&ptr3[1])->symbol->type)){
								jj=1;
							}
						}

					/* No clause coming from negated has been processed yet. */
					} else {
						jj=0;
					}
				}

				/* Clause must be flagged as kept. Flag the clause */
				/* and all its predicates and functions. Also indicate */
				/* that main loop must be repeated. */
				if (jj){
					ptr1->flags|=KEPTCLAUSE;
					for (ptr3=&ptr1->part2.bin->formula[0];ptr3[0]!=UNITEND;ptr3=NextItem(ptr3,IMMED)){
						if ((PREDICATE|FUNCTION)&ptr3[0]){
							((symbol *)&ptr3[1])->symbol->type|=RELEVANT;
						}
					}
					ii=1;

				/* Clause must not be flagged as kept. Set ptr2 if not already set. */
				} else {
					if (ptr2==NULL){
						ptr2=ptr1;
					}
				}
			}
		}

		/* Check timeout or number of iterations. */
		if (ppthrshld>0.0){
			if (prproflags&CLFILTERINGITERLIMIT){
				if (nn>=ppthrshld){
					if (ii){
						prproflags|=CLFILTERINGLIMITREACHED;
					}
					break;
				}
			} else {
				gettimeofday(&ctime,NULL);
				timedif=ctime.tv_sec-sttime.tv_sec
						+(ctime.tv_usec-sttime.tv_usec)/1000000.0;
				if (timedif>ppthrshld){
					if (ii){
						prproflags|=CLFILTERINGLIMITREACHED;
					}
					break;
				}
			}
		}
	}

	/* Delete non flagged clauses. Unlink() function cannot */
	/* return NOMEMORY because all the clauses are unprocessed. */
	/* None of the clauses have equality data nor asserts so */
	/* there is no need to free that storage. */
	if ((0==(prproflags&CLFILTERINGLIMITREACHED))||(0==(prproflags&CLFILTERINGRESIGN))){
		for (ptr1=mainkb->frstunproc,kk=0;ptr1!=NULL;ptr1=ptr2){
			ptr2=ptr1->next;
			if (0==(KEPTCLAUSE&ptr1->flags)){
				Unlink(ptr1,num_wkbs);
				MYFREE(ptr1->part2.bin);
				MYFREE(ptr1);
				kk++;
			} else {
				ptr1->flags&=~(KEPTCLAUSE);
			}
			if (kk){
				prproflags|=CLFILTERINGPRUNNED;
			}
		}
	} else {
		kk=0;
	}

	return(kk);
} /* GlobalPreprocess */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Print problem type flags)  OCJ
 *
 *    This function prints the problem type flags in pbmflags global
 *    variable. See global.h for the flags in pbmflags.
 *
 *    In order to do so the problem must be pre-processed.
 *
 *    All main KB (and therefore all KB's data) will be initialized
 *    at the end of the execution of this function.
 *
 *
 *  ARGUMENTS:
 *
 *    None.
 *
 *  RETURNS:
 *
 *    None.
 *
 *--------------------------------------------------------------*/
void GoType(void){
	auto char *uniteq,*unitcl,*horn,*ground,*eq;         /* To store "Y" or "N" for the different simple problem types */
	auto int32_t ii;                                     /* Auxiliary */

	/* Start time measurement. */
	procctl.start=clock();
	gettimeofday(&mainkb->time,NULL);

	/* No KB's available. */
	if (mainkb==NULL){
		printf("Cannot process: No KB's available\n");
		return;
	}

	/* Main KB is empty. */
	if (mainkb->nformulas==0){
		printf("Main KB has no input formulas to process.\n");
		return;
	}

	/* Perform pre-processing. */
	ii=Preprocess(0);
	if (ii){

		/* Print error information. */
		if (ii==NOMEMORY){
			printf("Not enough memory.\n");
		} else {
			printf("Process timeout\n");
		}
		printf("      UNIT_EQUALITY UNIT_CLAUSE HORN GROUND NONUNIT_EQU PBM_TYPE\n");
		printf("====>      ?             ?       ?      ?         ?        ?\n");
		FreeBinFormulas(num_wkbs);
		if (NOMEMORY==Initialize_KB(num_wkbs)){
			printf("No memory available, KB not initialized\n");
		}
		glblstats.elapsed_time+=mainkb->prstats.elapsed_time;
		ResetStats(&mainkb->prstats);
		procctl.status=STOPPED;
		tot_pbms++;

		/* Stop time measurement, collect CPU time and return. */
		procctl.end=clock();
		tot_cpu_time+=((double)(procctl.end-procctl.start))/CLOCKS_PER_SEC;
		return;
	}

	/* Display results. */
	if (pbmflags&UNITEQUALITY){
		uniteq="Y";
	} else {
		uniteq="N";
	}
	if (pbmflags&UNITCLAUSE){
		unitcl="Y";
	} else {
		unitcl="N";
	}
	if (pbmflags&HORN){
		horn="Y";
	} else {
		horn="N";
	}
	if (pbmflags&GROUND){
		ground="Y";
	} else {
		ground="N";
	}
	if (pbmflags&NONUNITEQUALITY){
		eq="Y";
	} else {
		eq="N";
	}
	printf("      UNIT_EQUALITY UNIT_CLAUSE HORN GROUND NONUNIT_EQU PBM_TYPE\n");
	printf("====>      %s             %s       %s      %s         %s        %d\n",
			uniteq,unitcl,horn,ground,eq,1+GetPbmTypeNumber());

	/* Collect memory use information. */
	if (glblstats.memory<allcdmemory){
		glblstats.memory=allcdmemory;
	}

	/* Collect inference statistics and initialize main KB. */
	AddInferenceStats(&mainkb->prstats,&glblstats);
	FreeBinFormulas(num_wkbs);
	if (NOMEMORY==Initialize_KB(num_wkbs)){
		printf("No memory available, KB not initialized\n");
	}
	glblstats.elapsed_time+=mainkb->prstats.elapsed_time;
	ResetStats(&mainkb->prstats);
	tot_pbms++;

	/* Stop time measurement, collect CPU time and return. */
	procctl.end=clock();
	tot_cpu_time+=((double)(procctl.end-procctl.start))/CLOCKS_PER_SEC;

	return;
} /* GoType */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Convert problem formulas to CNF and print them in TPTP format)  OCJ
 *
 *    This function converts problem formulas to CNF and prints them
 *    in TPTP format.
 *
 *    In order to do so the problem must be pre-processed.
 *
 *    All main KB (and therefore all KB's data) will be initialized
 *    at the end of the execution of this function.
 *
 *
 *  ARGUMENTS:
 *
 *    None.
 *
 *  RETURNS:
 *
 *    None.
 *
 *--------------------------------------------------------------*/
void Clausify(void){
	auto int32_t ii;                                     /* Auxiliary */

	/* Start time measurement. */
	procctl.start=clock();
	gettimeofday(&mainkb->time,NULL);

	/* No KB's available. */
	if (mainkb==NULL){
		printf("Cannot process: No KB's available\n");
		return;
	}

	/* Main KB is empty. */
	if (mainkb->nformulas==0){
		printf("Main KB has no input formulas to process.\n");
		return;
	}

	/* Perform pre-processing. */
	ii=Preprocess(2);
	if (ii){
		if (ii==NOMEMORY){
			printf("Not enough memory.\n");
		} else {
			printf("Process timeout\n");
		}
		procctl.end=clock();
		tot_cpu_time+=((double)(procctl.end-procctl.start))/CLOCKS_PER_SEC;
		gettimeofday(&mainkb->endtime,NULL);
		glblstats.elapsed_time+=(mainkb->endtime.tv_sec-mainkb->time.tv_sec
				+(mainkb->endtime.tv_usec-mainkb->time.tv_usec)/1000000.0);
		FreeBinFormulas(num_wkbs);
		if (NOMEMORY==Initialize_KB(num_wkbs)){
			printf("No memory available, KB not initialized\n");
		}
		ResetStats(&mainkb->prstats);
		procctl.status=STOPPED;
		tot_pbms++;
		return;
	}

	/* Display results. */
	if (NOMEMORY==PrintClausifiedFormulas()){
		printf("No memory available, printed information is not complete.\n");
	}

	/* Collect memory use information. */
	if (glblstats.memory<allcdmemory){
		glblstats.memory=allcdmemory;
	}

	/* Collect inference statistics and initialize main KB. */
	AddInferenceStats(&mainkb->prstats,&glblstats);
	FreeBinFormulas(num_wkbs);
	if (NOMEMORY==Initialize_KB(num_wkbs)){
		printf("No memory available, KB not initialized\n");
	}
	ResetStats(&mainkb->prstats);
	tot_pbms++;

	/* Stop time measurement and collect elapsed and CPU times. */
	procctl.end=clock();
	tot_cpu_time+=((double)(procctl.end-procctl.start))/CLOCKS_PER_SEC;
	gettimeofday(&mainkb->endtime,NULL);
	glblstats.elapsed_time+=(mainkb->endtime.tv_sec-mainkb->time.tv_sec
			+(mainkb->endtime.tv_usec-mainkb->time.tv_usec)/1000000.0);

	return;
} /* Clausify */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Select strategies for current problem)  OCJ
 *
 *    This function selects the strategies to be used for the current
 *    problem being solved. The selected strategies are set in the
 *    strategies[] global variable. Each strategy selected is a 16 bit
 *    unsigned integer with the values for each type of adjustable parameter.
 *    See definitions of adjustable parameters in global.h.
 *
 *    The best substrategies for each type of problem are stored in
 *    pbmtypestrats[ll][mm][nn] global variable, where:
 *    ll: 0 for strategies without learning, 1 for strategies with learning.
 *    mm: This is problem type index.
 *    nn: This is the index to the strategy itself.
 *
 *    The number of strategies are stored in nbpbmtypestr[ll][mm] global
 *    variable, where:
 *    ll: 0 for strategies without learning, 1 for strategies with learning.
 *    mm: This is problem type index, same as for pbmtypestrats[mm][nn].
 *
 *    This function selects more strategies than necessary in principle.
 *    This is done to deal with strategy runs that end with no solution
 *    in less time than available due to strategy incompleteness. Of course
 *    the number of possible strategies is limited by the number of active
 *    cores and the number of time slices.
 *
 *  ARGUMENTS:
 *
 *    None.
 *
 *  RETURNS:
 *
 *    0 if OK or NOMEMORY.
 *
 *
 *--------------------------------------------------------------*/
int32_t InitStrategies(void){
	auto int32_t sel_strats;                             /* Number of currently selected strategies */
	auto int8_t *usedstrats;                             /* Flags for used strategies. */
	auto int32_t ii,jj,kk;                               /* Auxiliary */
	auto uint16_t mm;                                    /* Auxiliary */

	/* Calculate problem type number and initialize nextstrat and nbstrats. */
	/* Set jj=1 if learning is enabled, 0 otherwise. */
	pbmtype=GetPbmTypeNumber();
	nextstrat=0;
	jj=(NULL==learnfrom?0:1);
	if ((num_wkbs==1)&&(timeslices==1)){
		nbstrats=1;
	} else {
		nbstrats=num_wkbs*(timeslices+BKSTRATEGIESPERTHREAD);
		if (nbstrats>MAXSTRATEGIES){
			nbstrats=MAXSTRATEGIES;
		}
	}

	/* Get the canonical strategies for the problem. */
	for (sel_strats=0;(sel_strats<nbpbmtypestr[jj][pbmtype])&&(sel_strats<nbstrats);
			sel_strats++){
		strategies[sel_strats]=StratNb2StratFlags(pbmtypestrats[jj][pbmtype][sel_strats]);
	}

	/* There are user parameter settings. */
	if (usrparam_mask){

		/* Check if user parameter settings match any of the selected strategies. */
		for (kk=0,ii=1;kk<sel_strats;kk++){
			if (usrparam==(strategies[kk]&usrparam_mask)){
				ii=0;
				break;
			}
		}

		/* If none of the selected strategies match the user parameter settings */
		/* then shift one position the selected strategies and combine the */
		/* user parameter settings with the first selected strategy. */
		if (ii){
			for (kk=(sel_strats==nbstrats?sel_strats-1:sel_strats);kk>0;kk--){
				strategies[kk]=strategies[kk-1];
			}
			strategies[0]=usrparam|(strategies[0]&(~usrparam_mask));
			if (sel_strats<nbstrats){
				sel_strats++;
			}
		}
	} else {
		ii=0;
	}

	/* More strategies must be selected. */
	if (sel_strats<nbstrats){

		/* There is enough memory for used strategies flags.*/
		if (NULL!=(usedstrats=mallocmn(MAXSTRATEGIES))){

			/* Set flags of used strategies without user specified strategy. */
			memset(usedstrats,0,MAXSTRATEGIES);
			for (kk=ii;kk<sel_strats;kk++){
				usedstrats[pbmtypestrats[jj][pbmtype][kk]]=1;
			}

			/* Add additional strategies. The additional strategies are selected */
			/* from non used strategies without / with learning alternatively. */
			for (kk=0;sel_strats<nbstrats;kk++){
				if (0==usedstrats[kk]){
					mm=StratNb2StratFlags(kk);
					if (mm!=strategies[0]){
						strategies[sel_strats]=mm;
						sel_strats++;
					}
				}
				if ((NULL!=learnfrom)&&(0==usedstrats[kk+(MAXSTRATEGIES/2)])&&(sel_strats<nbstrats)){
					mm=StratNb2StratFlags(kk+(MAXSTRATEGIES/2));
					if (mm!=strategies[0]){
						strategies[sel_strats]=mm;
						sel_strats++;
					}
				}
			}

			/* Free allocated memory. */
			MYFREE(usedstrats);

		/* Not enough memory. */
		} else {
			return(NOMEMORY);
		}
	}

	return(0);
} /* InitStrategies */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Calculate problem type number)  OCJ
 *
 *    This function calculates a problem type number according to
 *    the following:
 *    LEGEND:
 *    ------
 *    UE: Problem has some positive unit equalities.
 *    NUE: Problem has some equalities none of which are positive unit
 *         equalities.
 *    UC: All clauses in problem are unit clauses.
 *    H: Problem is horn, that is, all clauses contain at most only
 *       one positive literal.
 *    G: Problem is ground, there are no variables in the clauses.
 *
 *    PROBLEM TYPE NUMBER:
 *    -------------------
 *    0 -> UE&~UC&~H&~G
 *    1 -> UE&~UC&H
 *    2 -> UE&~UC&~H&G
 *    3 -> UE&UC
 *    4 -> NUE&~UC&~H y NUE&UC
 *    5 -> NUE&~UC&H
 *    6 -> ~UE&~NUE&~UC&~H y ~UE&~NUE&UC
 *    7 -> ~UE&~NUE&~UC&H
 *
 *  ARGUMENTS:
 *
 *    None.
 *
 *  RETURNS:
 *
 *    Problem type number.
 *
 *
 *--------------------------------------------------------------*/
int32_t GetPbmTypeNumber(void){

	/* Calculate problem type number. */
	/* 0 -> UE&~UC&~H&~G */
	/* 1 -> UE&~UC&H */
	/* 2 -> UE&~UC&~H&G */
	/* 3 -> UE&UC */
	/* 4 -> NUE&~UC&~H y NUE&UC */
	/* 5 -> NUE&~UC&H */
	/* 6 -> ~UE&~NUE&~UC&~H y ~UE&~NUE&UC */
	/* 7 -> ~UE&~NUE&~UC&H */
	switch (pbmflags&(UNITEQUALITY|NONUNITEQUALITY|UNITCLAUSE|HORN|GROUND)){
		case UNITEQUALITY:
			pbmtype=0;
			break;
		case UNITEQUALITY|HORN|GROUND:
		case UNITEQUALITY|HORN:
			pbmtype=1;
			break;
		case UNITEQUALITY|GROUND:
			pbmtype=2;
			break;
		case UNITEQUALITY|UNITCLAUSE:
		case UNITEQUALITY|UNITCLAUSE|HORN:
		case UNITEQUALITY|UNITCLAUSE|GROUND:
		case UNITEQUALITY|UNITCLAUSE|HORN|GROUND:
			pbmtype=3;
			break;
		case NONUNITEQUALITY:
		case NONUNITEQUALITY|GROUND:
		case NONUNITEQUALITY|UNITCLAUSE:
		case NONUNITEQUALITY|UNITCLAUSE|HORN:
		case NONUNITEQUALITY|UNITCLAUSE|GROUND:
		case NONUNITEQUALITY|UNITCLAUSE|HORN|GROUND:
			pbmtype=4;
			break;
		case NONUNITEQUALITY|HORN:
		case NONUNITEQUALITY|HORN|GROUND:
			pbmtype=5;
			break;
		case 0:
		case GROUND:
		case UNITCLAUSE:
		case UNITCLAUSE|HORN:
		case UNITCLAUSE|GROUND:
		case UNITCLAUSE|HORN|GROUND:
			pbmtype=6;
			break;
		/* The following cases are redundant. They are left just for documentation. */
		/* The compiler optimization will remove the redundance. */
		case HORN:
		case HORN|GROUND:
		default:
			pbmtype=7;
			break;
	}
	/* The following is the same as above written in terms of conditionals. */
	/* It is commented out as documentation and clarification. */
	/*if (pbmflags&UNITEQUALITY){
		if (0==(pbmflags&UNITCLAUSE)){
			if (pbmflags&HORN){
				pbmtype=1;
			} else if (pbmflags&GROUND){
				pbmtype=2;
			} else {
				pbmtype=0;
			}
		} else {
			pbmtype=3;
		}
	} else if (pbmflags&NONUNITEQUALITY){
		if ((0==(pbmflags&UNITCLAUSE))&&(pbmflags&HORN)){
			pbmtype=5;
		} else {
			pbmtype=4;
		}
	} else if ((0==(pbmflags&UNITCLAUSE))&&(pbmflags&HORN)){
		pbmtype=7;
	} else {
		pbmtype=6;
	}*/

	return(pbmtype);
} /* GetPbmTypeNumber */

/*--------------------------------------------------------------
 *
 *  DESCRIPTION: (Convert a strategy number to strategy flags)  OCJ
 *
 *    This function converts an integer with the strategy number to
 *    a 16-bytes unsigned integer with strategy flags.
 *
 *
 *  ARGUMENTS:
 *
 *    stratnb: Strategy number.
 *
 *  RETURNS:
 *
 *    16-bit integer with strategy flags.
 *
 *--------------------------------------------------------------*/
uint16_t StratNb2StratFlags(int32_t stratnb){
	auto uint16_t ii;                                    /* Auxiliary */

	/* Set learning. */
	if (stratnb>=(MAXSTRATEGIES/2)){
		ii=LEARN_P;
		stratnb-=(MAXSTRATEGIES/2);
	} else {
		ii=0;
	}

	/* Set algorithm and maximum weight option. */
	if (stratnb<(MAXSTRATEGIES/3)){
		if (stratnb&1){
			ii|=(DISCOUNT_P|LRSOTT_P);
		} else {
			ii|=(OTTER_P|LRSOTT_P);
		}
		stratnb/=2;
	} else {
		ii|=(OTTER_P|LRSOFF_P);
		stratnb-=(MAXSTRATEGIES/3);
	}

	/* Set literal selection options. */
	switch (stratnb%5){
		case 0:
			ii|=SINGLENEG_P;
			break;
		case 1:
			ii|=MULTINEG_P;
			break;
		case 2:
			ii|=MAXIMAL_P;
			break;
		case 3:
			ii|=SINGLEPOS_P;
			break;
		case 4:
			ii|=MULTIPOS_P;
			break;
	}

	/* Set weight factors option for clause weight calculation. */
	if ((stratnb/5)%2){
		ii|=WFACT2_P;
	} else {
		ii|=WFACT1_P;
	}

	/* Set equivalent factoring inference enabled/disabled. */
	if ((stratnb/10)%2){
		ii|=EQFACTOFF_P;
	} else {
		ii|=EQFACTON_P;
	}

	/* Set factoring inferences enabled/disabled. */
	if ((stratnb/20)%2){
		ii|=FACTOFF_P;
	} else {
		ii|=FACTON_P;
	}

	/* Set term ordering. */
	if ((stratnb/40)%2){
		ii|=TRMORDNR_P;
	} else {
		ii|=TRMORDSTD_P;
	}

	/* Set literal ordering. */
	switch ((stratnb/80)%3){
		case 0:
			ii|=LITORDSTD_P;
			break;
		case 1:
			ii|=LITORDNR_P;
			break;
		case 2:
			ii|=LITORDLEX_P;
			break;
	}

	/* Set function weight option. */
	if ((stratnb/240)%2){
		ii|=UNIFORM_P;
	} else {
		ii|=ARITY_P;
	}
	return(ii);
} /* StratNb2StratFlags */
